import os
import sys
import time
import logging


def is_tty():
    # Look if we are in a tty or not
    if hasattr(sys.stdout, 'isatty'):
        return sys.stdout.isatty()
    return False

if is_tty():
    # Try to load the terminal color. Won't work under python 2.4
    try:
        from kunai.termcolor import cprint
    except (SyntaxError, ImportError), exp:
        # Outch can't import a cprint, do a simple print
        def cprint(s, color='', end='\n'):
            if end == '':
                print s,
            else:
                print s

# Ok it's a daemon mode, if so, just print
else:
    def cprint(s, color='', end='\n'):
        if end == '':
            print s,
        else:
            print s


            
            
            
class Logger(object):
    def __init__(self):
        self.data_dir = ''
        self.log_file = None
        self.name = ''
        self.logs = {}
        self.level = logging.INFO
        self.linkify_methods()

        # We will keep last 20 errors
        self.last_errors_stack_size = 20
        self.last_errors_stack = {'DEBUG':[], 'WARNING':[], 'INFO':[], 'ERROR':[]}
        

    def linkify_methods(self):
        methods = {'DEBUG': self.do_debug, 'WARNING': self.do_warning, 'INFO': self.do_info, 'ERROR':self.do_error}
        for (s, m) in methods.iteritems():
            level = getattr(logging, s)
            # If the level is enough, link it
            if level >= self.level:
                setattr(self, s.lower(), m)
            else:
                setattr(self, s.lower(), self.do_null)
        
        
    def load(self, data_dir, name):
        self.name = name
        self.data_dir = data_dir
        self.log_file = open(os.path.join(self.data_dir, 'daemon.log'), 'a')        


    def setLevel(self, s):
        try:
            level = getattr(logging , s.upper())
            if not isinstance(level, int):
                raise AttributeError
            self.level = level
        except AttributeError:
            self.error('Invalid logging level configuration %s' % s)
            return
        
        self.linkify_methods()
            

    def get_errors(self):
        return self.last_errors_stack
        

    def log(self, *args, **kwargs ):
       name = self.name
       now = int(time.time())
       s = '%s [%d]: %s' % (name, now, ' '.join([ str(s) for s in args]))
       if 'color' in kwargs:
           cprint(s, color=kwargs['color'])
       else:
           print(s)
       stack = kwargs.get('stack', False)

       # Not a perf problems as it's just for errors and a limited size
       if stack:
           self.last_errors_stack[stack].append(s)
           # And keep only the last 20 ones for example
           self.last_errors_stack[stack] = self.last_errors_stack[stack][-self.last_errors_stack_size:]

       # if no data_dir, we cannot save anything...
       if self.data_dir == '':
           return
       part = kwargs.get('part', '')
       if part == '':
           if self.log_file is not None:
               self.log_file.write(s+'\n')
       else:
           f = self.logs.get(part, None)
           if f is None:
               f = open(os.path.join(self.data_dir, '%s.log' % part), 'a')
               self.logs[part] = f
           f.write(s+'\n')

    
    def do_debug(self, *args, **kwargs):
        self.log(*args, color='magenta', **kwargs)
        

    def do_info(self, *args, **kwargs):
        self.log(*args, color='blue', **kwargs)

        
    def do_warning(self, *args, **kwargs):
        self.log(*args, color='yellow', stack='WARNING', **kwargs)        

        
    def do_error(self, *args,  **kwargs):
        self.log(*args, color='red', stack='ERROR',  **kwargs)

    
    def do_null(self, *args,  **kwargs):
        pass

        
logger = Logger()



#!/usr/bin/env python
# Copyright (C) 2014:
#    Gabes Jean, naparuba@gmail.com

import copy
import time
import threading
from kunai_test import *

from kunai.encrypter import encrypter



class TestEncrypter(KunaiTest):
    def setUp(self):
        encrypter.load('NTdiN2NlNmE4NTViMTFlNA==')


    def test_encryption(self):
        orig_test = 'Hi I am Alice'
        bobread = encrypter.encrypt(orig_test)
        print bobread
        clear = encrypter.decrypt(bobread)
        print clear
        self.assert_(clear == orig_test)

if __name__ == '__main__':
    unittest.main()



import httplib # Used only for handling httplib.HTTPException (case #26701)
import os
import platform
import re
import urllib
import urllib2
import traceback
import time
from StringIO import StringIO


from kunai.log import logger
from kunai.collector import Collector


class RabbitMQ(Collector):
    def launch(self):
        logger.debug('getRabbitMQStatus: start')

        if 'rabbitMQStatusUrl' not in self.config or \
                    'rabbitMQUser' not in self.config or \
                    'rabbitMQPass' not in self.config or \
            self.config['rabbitMQStatusUrl'] == 'http://www.example.com:55672/json':

            logger.debug('getRabbitMQStatus: config not set')
            return False

        logger.debug('getRabbitMQStatus: config set')

        try:
            logger.debug('getRabbitMQStatus: attempting authentication setup')

            manager = urllib2.HTTPPasswordMgrWithDefaultRealm()
            manager.add_password(None, self.config['rabbitMQStatusUrl'], self.config['rabbitMQUser'], self.config['rabbitMQPass'])
            handler = urllib2.HTTPBasicAuthHandler(manager)
            opener = urllib2.build_opener(handler)
            urllib2.install_opener(opener)

            logger.debug('getRabbitMQStatus: attempting urlopen')
            req = urllib2.Request(self.config['rabbitMQStatusUrl'], None, headers)

            # Do the request, log any errors
            request = urllib2.urlopen(req)
            response = request.read()

        except urllib2.HTTPError, e:
            logger.error('Unable to get RabbitMQ status - HTTPError = %s', e)
            return False

        except urllib2.URLError, e:
            logger.error('Unable to get RabbitMQ status - URLError = %s', e)
            return False

        except httplib.HTTPException, e:
            logger.error('Unable to get RabbitMQ status - HTTPException = %s', e)
            return False

        except Exception, e:
            logger.error('Unable to get RabbitMQ status - Exception = %s', traceback.format_exc())
            return False

        try:                
            if int(pythonVersion[1]) >= 6:
                logger.debug('getRabbitMQStatus: json read')
                status = json.loads(response)

            else:
                logger.debug('getRabbitMQStatus: minjson read')
                status = minjson.safeRead(response)

            logger.debug(status)

            if 'connections' not in status:
                # We are probably using the newer RabbitMQ 2.x status plugin, so try to parse that instead.
                status = {}
                connections = {}
                queues = {}
                logger.debug('getRabbitMQStatus: using 2.x management plugin data')
                import urlparse

                split_url = urlparse.urlsplit(self.config['rabbitMQStatusUrl'])

                # Connections
                url = split_url[0] + '://' + split_url[1] + '/api/connections'
                logger.debug('getRabbitMQStatus: attempting urlopen on %s', url)
                manager.add_password(None, url, self.config['rabbitMQUser'], self.config['rabbitMQPass'])
                req = urllib2.Request(url, None, headers)
                # Do the request, log any errors
                request = urllib2.urlopen(req)
                response = request.read()

                if int(pythonVersion[1]) >= 6:
                    logger.debug('getRabbitMQStatus: connections json read')
                    connections = json.loads(response)
                else:
                    logger.debug('getRabbitMQStatus: connections minjson read')
                    connections = minjson.safeRead(response)

                status['connections'] = len(connections)
                logger.debug('getRabbitMQStatus: connections = %s', status['connections'])

                # Queues
                url = split_url[0] + '://' + split_url[1] + '/api/queues'
                logger.debug('getRabbitMQStatus: attempting urlopen on %s', url)
                manager.add_password(None, url, self.config['rabbitMQUser'], self.config['rabbitMQPass'])
                req = urllib2.Request(url, None, headers)
                # Do the request, log any errors
                request = urllib2.urlopen(req)
                response = request.read()

                if int(pythonVersion[1]) >= 6:
                    logger.debug('getRabbitMQStatus: queues json read')
                    queues = json.loads(response)
                else:
                    logger.debug('getRabbitMQStatus: queues minjson read')
                    queues = minjson.safeRead(response)

                status['queues'] = queues
                logger.debug(status['queues'])

        except Exception, e:
            logger.error('Unable to load RabbitMQ status JSON - Exception = %s', traceback.format_exc())
            return False

        logger.debug('getRabbitMQStatus: completed, returning')

        # Fix for queues with the same name (case 32788)
        for queue in status.get('queues', []):
            vhost = queue.get('vhost', '/')
            if vhost == '/':
                continue

            queue['name'] = '%s/%s' % (vhost, queue['name'])

        return status



import os
import json
import time
import leveldb
import threading
import shutil

from kunai.stats import STATS
from kunai.log import logger
from kunai.threadmgr import threader
from kunai.now import NOW

# This class manage the ttl entries for each key with a ttl. Each is with a 1hour precisionidx key that we saved in the master db
# but with keeping a database by hour about the key for the housekeeping
class TTLDatabase(object):
   def __init__(self, kv, ttldb_dir):
      self.lock = threading.RLock()
      self.dbs = {}
      self.db_cache_size = 100
      self.kv = kv
      self.ttldb_dir = ttldb_dir
      if not os.path.exists(self.ttldb_dir):
         os.mkdir(self.ttldb_dir)
      # Launch a thread that will look once a minute the old entries
      threader.create_and_launch(self.ttl_cleaning_thread, name='ttl-cleaning-thread')


   # Load the hour ttl/H base where we will save all master
   # key for the H hour
   def get_ttl_db(self, h):
      cdb = self.dbs.get(h, None)
      # If missing, look to load it but with a lock to be sure we load it only once
      if cdb is None:
         STATS.incr('ttl-db-cache-miss',1)
         with self.lock:
            # Maybe during the lock get one other thread succedd in getting the cdb
            if not h in self.dbs:
               # Ok really load it, but no more than self.db_cache_size
               # databases (number of open files can increase quickly)
               if len(self.dbs) > self.db_cache_size:
                  ttodrop = self.dbs.keys()[0]
                  del self.dbs[ttodrop]
               _t = time.time()
               cdb = leveldb.LevelDB(os.path.join(self.ttldb_dir, '%d' % h))
               STATS.incr('ttl-db-open', time.time() - _t)
               self.dbs[h] = cdb
            # Ok a malicious thread just go before us, good :)
            else:
               cdb = self.dbs[h]
      # We alrady got it, thanks cache
      else:
         STATS.incr('ttl-db-cache-hit',1)
      return cdb


   # Save a key in the good idx minute database
   def set_ttl(self, key, ttl_t):
      # keep keys saved by hour in the future
      ttl_t = divmod(ttl_t, 3600)[0]*3600

      cdb = self.get_ttl_db(ttl_t)
      logger.debug("TTL save", key, "with ttl", ttl_t, "in", cdb, part='kv')
      cdb.Put(key, '')

      
   # We already droped all entry in a db, so drop it from our cache
   def drop_db(self, h):
      # now demove the database
      with self.lock:
         try:
            del self.dbs[h]
         except IndexError: # if not there, not a problem...
            pass

      # And remove the files of this database
      p = os.path.join(self.ttldb_dir, '%d' % h)
      logger.log("Deleting ttl database tree", p, part='kv')
      shutil.rmtree(p, ignore_errors=True)
      

   # Look at the available dbs and clean all olds dbs that time are lower 
   # than current hour
   def clean_old(self):
      logger.debug("TTL clean old", part='kv')
      now = NOW.now + 3600
      h = divmod(now, 3600)[0]*3600
      # Look at the databses directory that have the hour time set
      subdirs = os.listdir(self.ttldb_dir)

      for d in subdirs:
         try:
            bhour = int(d)
         except ValueError: # who add a dir that is not a int here...
            continue
         # Is the hour available for cleaning?
         if bhour < h:
            logger.log("TTL bhour is too low!", bhour, part='kv')
            # take the database and dump all keys in it
            cdb = self.get_ttl_db(bhour)
            to_del = cdb.RangeIter()
            # Now ask the cluster to delete the key, whatever it is
            for (k,v) in to_del:
               self.kv.delete(k)
               
            # now we clean all old entries, remove the idx database
            self.drop_db(bhour)


   # Thread that will manage the delete of the ttld-die key
   def ttl_cleaning_thread(self):
      while True:
         time.sleep(5)
         self.clean_old()


# Main KV backend. Reply on a local leveldb database. It's up to the
# cluster to know if we should manage a key or not, if someone give us it,
# we save it :)
class KVBackend:
   def __init__(self, data_dir):
      self.data_dir = data_dir
      self.db_dir = os.path.join(data_dir, 'kv')
      self.db = leveldb.LevelDB(self.db_dir)
      self.ttldb = TTLDatabase(self, os.path.join(data_dir, 'ttl'))

      self.update_db_time = 0
      self.update_db = None


   # We will open a file with the keys writen during a minute
   # so we can easily look at previous changed
   def get_update_db(self, t):
      cmin = divmod(t, 60)[0] * 60
      if cmin == self.update_db_time and self.update_db:
         #print "UPDATE DB CACHE HIT"
         return self.update_db
      else: # not the good time
         #print "UPDATE DB CACHE MISS"
         if self.update_db:
            print "FLUSINH PREVIOUS DB"
            t0 = time.time()
            self.update_db.flush()
            print "FLUSH TIME", time.time() - t0
            self.update_db.close()
         db_dir = os.path.join(self.data_dir, 'updates')
         db_path = os.path.join(db_dir, '%d.lst' % cmin)
         if not os.path.exists(db_dir):
            os.mkdir(db_dir)
         self.update_db = open(db_path, 'a')
         self.update_db_time = cmin
         return self.update_db


   # Raw get in our db for a key
   def get(self, key):
      try:
         t0 = time.time()
         v = self.db.Get(key)
         logger.debug("TIME kv get", time.time() - t0, part='kv')
         return v
      except KeyError:
         return None


   # Put a key/value in leveldb. Compute the meta
   # entry and increate the modify_index (+1) and modify_time too
   # If ttl is et (!=0) then add an entry in a TTL database
   def put(self, key, value, ttl=0):
      # manage the meta data for this entry 
      # like modification index
      metakey = '__meta/%s' % key
      try:
         metavalue = json.loads(self.db.Get(metakey))
      except KeyError:
         metavalue = {'modify_index':0, 'modify_time':0}

      metavalue['modify_index'] += 1
      mtime = NOW.now#int(time.time())
      metavalue['modify_time'] = mtime

      # Update our meta values
      self.put_meta(key, metavalue)

      # if we got a tll, compute the dead time, and set it
      if ttl>0:
         dead_t = NOW.now + ttl#int(time.time()) + ttl
         self.ttldb.set_ttl(key, dead_t)
      
      # also put an entry to the update_db
      f = self.get_update_db(mtime)
      t0 = time.time()
      f.write('%s\n' % key)
      #print "write", time.time() - t0
      
      t0 =time.time()
      # and in the end save the real data :)
      self.db.Put(key, value)
      #print "write db", time.time() - t0


   # Delete both leveldb and metadata entry
   def delete(self, key):
      try:
         self.db.Delete(key)
      except KeyError:
         pass
      # also delete the meta entry
      metakey = '__meta/%s' % key
      try:
         self.db.Delete(metakey)
      except KeyError:
         pass


   # Get a json dump of a metadata entry
   def get_meta(self, key):
      metakey = '__meta/%s' % key
      v = self.get(metakey)
      if v is None:
         return v
      return json.loads(v)

      
   # Save a metadata entry in json
   def put_meta(self, key, meta):
      metakey = '__meta/%s' % key
      metadata = meta
      if isinstance(meta, dict):
         metadata = json.dumps(meta)
      self.db.Put(metakey, metadata)
   

   # Look at meta entries for data that changed since t
   def changed_since(self, t):
      # Lookup all __meta keys
      _all = list(self.db.RangeIter(key_from = '__meta', key_to='__n'))

      r = []
      for (mkey, metaraw) in _all:
         meta = json.loads(metaraw)
         
         # maybe this key is too old to be interesting
         if meta['modify_time'] <= t:
            continue
         
         ukey = mkey[len('__meta')+1:]
         try:
            v = self.db.Get(ukey)
         except KeyError: #should never be possible
            continue
         r.append( (ukey, v, meta) )
      return r
      

   # Try to merge distant data from others with meta entries
   # and only take the data that are the newest
   def do_merge(self, to_merge):
      for (ukey, v, meta) in to_merge:
         metakey = '__meta/%s' % ukey
         try:
            lmeta = json.loads(self.db.Get(metakey))
         except KeyError:
            continue
         # If the other mod_index is higer, we import it :)
         if meta['modify_index'] > lmeta['modify_index']:
            self.put_meta(ukey, meta)
            self.db.Put(ukey, v)
         else:
            pass
            



#!/usr/bin/env python
import json
import time
import sys
try:
    from docker import Client
except ImportError:
    Client = None

from kunai.stats import STATS
from kunai.log import logger
from kunai.threadmgr import threader
from kunai.now import NOW
from kunai.httpdaemon import route, response


def lower_dict(d):    
    nd = {}
    for k,v in d.iteritems():
        nk = k.lower()
        if isinstance(v, dict): # yes, it's recursive :)
            v = lower_dict(v)
        nd[nk] = v
    return nd


class DockerManager(object):
    def __init__(self):
        self.con = None
        self.containers = {}
        # We got an object, we can fill the http daemon part
        self.export_http()

        
    def launch(self):
        if not Client:
            logger.warning('Missing docker lib')
            return
        t = threader.create_and_launch(self.do_loop, name='docker-loop')


    def connect(self):
        if not self.con:
            try:
                self.con = Client(base_url='unix://var/run/docker.sock')
            except Exception, exp:
                logger.debug('Cannot connect to docker %s' % exp)
                self.con = None


    def load_container(self, _id):
        inspect = self.con.inspect_container(_id)
        c = lower_dict(inspect)
        logger.debug('LOADED NEW CONTAINER %s' % c)
        self.containers[_id] = c
        
    
    def load_containers(self):
        if not self.con:
            return
        conts = self.con.containers()
        for c in conts:
            _id = c.get('Id')
            self.load_container(_id)
            print "Container", self.containers[_id]
                
        
    def do_loop(self):
        self.connect()
        self.load_containers()
        while True:
            self.connect()
            if not self.con:
                time.sleep(1) # do not hammer the connexion
                continue
            # now manage events and lock on it
            evts = self.con.events() # can lock here
            for ev in evts:
                evdata = json.loads(ev)
                _id = evdata["id"]
                status = evdata["status"]
                if status in ("die", "stop"):
                    if _id in self.containers:
                        logger.debug('removing a container %s' % _id)
                        del self.containers[_id]
                    else:
                        logger.error('Asking to remove an unknow container? %s' % _id)
                elif status == 'start':
                    self.load_container(_id)
                else:
                    logger.debug('UNKNOWN EVENT IN DOCKER %s' % status)

                    
    # main method to export http interface. Must be in a method that got
    # a self entry
    def export_http(self):

        @route('/docker/')
        @route('/docker')
        def get_docker():
            response.content_type = 'application/json'
            return json.dumps(self.con is not None)
    

        @route('/docker/containers')
        @route('/docker/containers/')
        def get_containers():
            response.content_type = 'application/json'
            return json.dumps(self.containers.values())

        
        @route('/docker/containers/:_id')
        def get_container(_id):
            response.content_type = 'application/json'
            cont = self.containers.get(_id, None)
            return json.dumps(cont)
    

        @route('/docker/images')
        @route('/docker/images/')
        def get_images():
            response.content_type = 'application/json'
            if self.con is None:
                return json.dumps(None)
            imgs = self.con.images()
            r = [lower_dict(d) for d in imgs]
            return json.dumps(r)

        
        @route('/docker/images/:_id')
        def get_images(_id):
            response.content_type = 'application/json'
            if self.con is None:
                return json.dumps(None)
            imgs = self.con.images()
            for d in imgs:
                if d['Id'] == _id:
                    return json.dumps(lower_dict(d))
            return json.dumps(None)

                    

dockermgr = DockerManager()
                    



from kunai.bottle import route, run, request, abort, error, redirect, response, gserver
import kunai.bottle as bottle
bottle.debug(True)

from kunai.log import logger
from kunai.kv import KVBackend
from kunai.dnsquery import DNSQuery
from kunai.ts import TSListener
from kunai.wsocket import WebSocketBackend
from kunai.util import make_dir, copy_dir
from kunai.threadmgr import threader
from kunai.perfdata import PerfDatas
from kunai.now import NOW


# now singleton objects
from kunai.websocketmanager import websocketmgr
from kunai.broadcast import broadcaster


# We want the http daemon to be accessible from everywhere without issue
class EnableCors(object):
    name = 'enable_cors'
    api = 2
    
    def apply(self, fn, context):
        def _enable_cors(*args, **kwargs):
            # Set CORS headers
            response.headers['Access-Control-Allow-Origin'] = '*'
            response.headers['Access-Control-Allow-Methods'] = 'GET, POST, PUT, OPTIONS, DELETE, PATCH'
            response.headers['Access-Control-Allow-Headers'] = 'Origin, Accept, Content-Type, X-Requested-With, X-CSRF-Token, X-Shinken-Token'
            response.headers['Access-Control-Allow-Crendentials'] = 'true'
            if bottle.request.method != 'OPTIONS':
                # actual request; reply with the actual response
                return fn(*args, **kwargs)

        return _enable_cors

    


# This class is the http daemon main interface
# in a singleton mode so you can easily register new uri from other
# part of the code, mainly by adding new route to bottle
class HttpDaemon(object):
    def __init__(self):
        pass


    def run(self, addr, port, socket_path):
        # First enable cors on all our calls
        bapp = bottle.app()
        bapp.install(EnableCors())

        if socket_path:
            # Will lock for in this
            # warning: without the str() cherrypy is not happy with the value, do not emove it
            run(server='cherrypy', bind_addr=str(socket_path), numthreads=8) # not need for a lot of threads here
        else:
            # And this too but in another thread
            run(host=addr, port=port, server='cherrypy', numthreads=64)# 256?

            
    # Some default URI    
    @error(404)
    def err404(error):
        return ''

    
    @route('/')
    def slash():
        return 'OK'

        


httpdaemon = HttpDaemon()




import os
import leveldb
import time
import threading
import random
import shutil
import socket
import base64
import cPickle
import marshal
import json
import hashlib
SERIALIZER = cPickle


from kunai.stats import STATS
from kunai.log import logger
from kunai.threadmgr import threader
from kunai.now import NOW


# Will have to forward to graphit intead
class UDPSender(object):
   def __init__(self, clust):
      self.clust = clust
      
      
   def push_key(self, k, v, ttl=0):
      T0 = time.time()
      STATS.incr('ts.graphite.push-key', 1)
      v64 = base64.b64encode(v)
      logger.debug("PUSH KEY", k, "and value", len(v64))
      #self.clust.put_key(k, v64, allow_udp=True, ttl=ttl)
      self.clust.stack_put_key(k, v64, ttl=ttl)
      STATS.timer('ts.graphite.push-key', (time.time() - T0)*1000)
      return




class TSBackend(object):
   def __init__(self, usender, clust):
      #self.its = its
      self.usender = usender
      self.data = {}
      self.data_lock = threading.RLock()
      self.max_data_age = 5
      # Get a db with all our metrics listed
      # We will save all the metrics we have
      self.clust = clust
      self.db_dir = os.path.join(clust.data_dir, 'ts')
      self.db = leveldb.LevelDB(self.db_dir)
      # Ok start our inner thread
      self.launch_reaper_thread()


   # push a name but return if the name was already there or not
   def set_name_if_unset(self, key):
      try:
         self.db.Get(key, fill_cache=False)
      except KeyError:
         self.db.Put(key, '')
         logger.debug('TS propagating a new key', key)
         # now propagate the key to the other ts nodes
         self.clust.stack_new_ts_broadcast(key)
      return False


   # List all keys in our name db that match the begining of key
   def list_keys(self, key):
      if key:
         start = key
         l = chr(ord(key[-1]) + 1)
         end = key[:-1]+l
         print "LOOKUP from %s to %s" % (start, end)
         r = self.db.RangeIter(start, end, include_value=False, fill_cache=False)
      else: # Full scan? are you crazy?
         print "LOOKUP full scan"
         r = self.db.RangeIter(include_value=False, fill_cache=False)         
      return list(r)
      
      


   # We consume data and create a new data entry if need
   def add_value(self, t, key, v):
      # do nothing, will have to forward to graphite instead
      return

   # The reaper thread look at old minute objects that are not updated since long, and 
   # force to archive them
   def launch_reaper_thread(self):
       threader.create_and_launch(self.do_reaper_thread, name='TS-reaper-thread')
       

   def do_reaper_thread(self):
       while True:
          time.sleep(1) # no more need as we will forward ts
                   




class TSListener(object):
   def __init__(self, clust):
      self.clust = clust
      self.addr = '0.0.0.0'
      self.statsd_port = 8125
      self.graphite_port = 2003
      self.last_write = time.time()
      self.nb_data = 0
      self.stats_interval = 10

      # Do not step on your own foot...
      self.stats_lock = threading.RLock()
      
      # our main data structs
      self.gauges = {}
      self.timers = {}
      self.histograms = {}
      self.counters = {}
      
      # our helper objects
      #self.its = IdxTSDatabase(self.clust)
      self.usender = UDPSender(self.clust)
      self.tsb = TSBackend(self.usender, self.clust)
      
      # Now start our threads
      # STATSD
      threader.create_and_launch(self.launch_statsd_udp_listener, name='TSL_statsd_thread')
      threader.create_and_launch(self.launch_compute_stats_thread, name='TSC_thread')
      # GRAPHITE
      threader.create_and_launch(self.launch_graphite_udp_listener, name='TSL_graphite_udp_thread')
      threader.create_and_launch(self.launch_graphite_tcp_listener, name='TSL_graphite_tcp_thread')
      self.graphite_queue = []
      threader.create_and_launch(self.graphite_reaper, name='graphite-reaper-thread')
      

   def log(self, *args):
      logger.log(*args)
      

   # push a name but return if the name was already there or not
   def set_name_if_unset(self, key):
      return self.tsb.set_name_if_unset(key)


   # list all keys that start with key
   def list_keys(self, key):
      return self.tsb.list_keys(key)
   

   def to_best_int_float(self, val):
      try:
         i = int(float(val))
         f = float(val)
      except ValueError:
         return None
      # If the f is a .0 value,
      # best match is int
      if i == f:
         return i
      return f


   # The compute stats thread compute the STATSD values each X
   # seconds and push them into the classic TS part
   def launch_compute_stats_thread(self):
      while True:
         now = time.time()
         if now > self.last_write + self.stats_interval:
            self.compute_stats()
            self.last_write = now
         time.sleep(0.1)


   def compute_stats(self):
      now = int(time.time())
      logger.debug("Computing stats")
      names = []
      
      # First gauges, we take the data and put a void dict instead so the other thread can work now
      with self.stats_lock:
         gauges = self.gauges
         self.gauges = {}

      for mname in gauges:
         _sum, nb, _min, _max = gauges[mname]
         _avg = _sum / float(nb)
         key = 'stats.gauges.'+mname
         self.tsb.add_value(now, key, _avg)
         key = 'stats.gauges.'+mname+'.min'
         self.tsb.add_value(now, key, _min)
         key = 'stats.gauges.'+mname+'.max'
         self.tsb.add_value(now, key, _max)

      # Now counters
      with self.stats_lock:
         counters = self.counters
         self.counters = {}

      for mname in counters:
         cvalue, ccount = counters[mname]
         # count
         key = 'stats.gauges.'+mname+'.count'
         self.tsb.add_value(now, key, cvalue)
         # rate
         key = 'stats.gauges.'+mname+'.rate'
         self.tsb.add_value(now, key, cvalue / self.stats_interval)

      # Now timers, lot of funs :)
      with self.stats_lock:
         timers = self.timers
         self.timers = {}

      _t = time.time()
      for (mname, timer) in timers.iteritems():
         # We will need to compute the mean_99, count_99, upper_99, sum_99, sum_quares_99
         # but also std, upper, lower, count, count_ps, sum, sum_square, mean, median
         _t = time.time()
         npvalues = np.array(timer)
         # Mean
         mean = np.mean(npvalues)
         key = 'stats.timers.'+mname+'.mean'
         self.tsb.add_value(now, key, mean)
         
         # Upper 99th, percentile
         upper_99 = np.percentile(npvalues, 99)
         key = 'stats.timers.'+mname+'.upper_99'
         self.tsb.add_value(now, key, upper_99)
         
         # Sum 99
         sum_99 = npvalues[:(npvalues<upper_99).argmin()].sum()
         key = 'stats.timers.'+mname+'.sum_99'
         self.tsb.add_value(now, key, sum_99)

         # Standard deviation
         std = np.std(npvalues)
         key = 'stats.timers.'+mname+'.std'
         self.tsb.add_value(now, key, std)

         # Simple count
         count = len(timer)
         key = 'stats.timers.'+mname+'.count'
         self.tsb.add_value(now, key, count)
         
         # Sum of all
         _sum = np.sum(npvalues)
         key = 'stats.timers.'+mname+'.sum'
         self.tsb.add_value(now, key, _sum)

         # Median of all
         median = np.percentile(npvalues, 50)
         key = 'stats.timers.'+mname+'.median'
         self.tsb.add_value(now, key, median)

         # Upper of all
         upper = np.max(npvalues)
         key = 'stats.timers.'+mname+'.upper'
         self.tsb.add_value(now, key, upper)

         # Lower of all
         lower = np.min(npvalues)
         key = 'stats.timers.'+mname+'.lower'
         self.tsb.add_value(now, key, lower)
         

   # This is ht main STATSD UDP listener thread. Should not block and 
   # be as fast as possible
   def launch_statsd_udp_listener(self):
        self.udp_sock = socket.socket(socket.AF_INET, socket.SOCK_DGRAM) # UDP
        self.udp_sock.setsockopt(socket.SOL_SOCKET, socket.SO_RCVBUF, 1048576)
        self.udp_sock.setsockopt(socket.SOL_SOCKET, socket.SO_REUSEADDR, 1)
        self.log(self.udp_sock.getsockopt(socket.SOL_SOCKET, socket.SO_RCVBUF))
        self.udp_sock.bind((self.addr, self.statsd_port))
        self.log("TS UDP port open", self.statsd_port)
        self.log("UDP RCVBUF", self.udp_sock.getsockopt(socket.SOL_SOCKET, socket.SO_RCVBUF))
        while True:#not self.interrupted:
            try:
               data, addr = self.udp_sock.recvfrom(65535) # buffer size is 1024 bytes
            except socket.timeout: # loop until we got something
               continue
            
            self.log("UDP: received message:", data, addr)
            # No data? bail out :)
            if len(data) == 0:
                continue
            self.log("GETDATA", data)
            
            for line in data.splitlines():
               # avoid invalid lines
               if not '|' in line:
                  continue
               elts = line.split('|', 1)
               # invalid, no type in the right part
               if len(elts) == 1:
                  continue

               _name_value = elts[0].strip()
               # maybe it's an invalid name...
               if not ':' in _name_value:
                  continue
               _nvs = _name_value.split(':')
               if len(_nvs) != 2:
                  continue
               mname = _nvs[0].strip()
               
               # Two cases: it's for me or not
               hkey = hashlib.sha1(mname).hexdigest()
               ts_node_manager = self.clust.find_ts_node(hkey)
               # if it's me that manage this key, I add it in my backend
               if ts_node_manager != self.clust.uuid:
                  node = self.clust.nodes.get(ts_node_manager, None)
                  # threads are dangerous things...
                  if node is None:
                     continue
                  
                  # TODO: do bulk send of this, like for graphite
                  sock = socket.socket(socket.AF_INET, socket.SOCK_DGRAM)              
                  # do NOT use the node['port'], it's the internal communication, not the graphite one!
                  sock.sendto(line, (node['addr'], self.statsd_port))
                  sock.close()
                  continue

               # Here we are sure it's really for us, so manage it :)
               value = self.to_best_int_float(_nvs[1].strip())
               if not mname or value is None:
                  continue

               # Look at the type of the data
               _type = elts[1].strip()
               if len(_type) == 0:
                  continue
               
               ## Gauge: <metric name>:<value>|g
               elif _type == 'g':
                  self.nb_data += 1
                  self.log('GAUGE', mname, value)
                  with self.stats_lock:
                     gentry = self.gauges.get(mname, None)
                     if gentry is None:
                        # sum, nb, min, max
                        gentry = (0.0, 0, None, None)
                     _sum, nb, _min, _max = gentry
                     _sum += value
                     nb += 1
                     if _min is None or value < _min:
                        _min = value
                     if _max is None or value > _max:
                        _max = value
                     self.gauges[mname] = (_sum, nb, _min, _max)
                     self.log('NEW GAUGE', mname, self.gauges[mname])
                     
               ## Timers: <metric name>:<value>|ms
               ## But also 
               ## Histograms: <metric name>:<value>|h
               elif _type == 'ms' or _type == 'h':
                  self.log('timers', mname, value)
                  # TODO: avoit the SET each time
                  timer = self.timers.get(mname, [])
                  timer.append(value)
                  self.timers[mname] = timer
               ## Counters: <metric name>:<value>|c[|@<sample rate>]
               elif _type == 'c':
                  self.nb_data += 1
                  self.log('COUNTER', mname, value, "rate", 1)
                  with self.stats_lock:
                     cvalue, ccount  = self.counters.get(mname, (0,0))
                     self.counters[mname] = (cvalue + value, ccount + 1)
                     self.log('NEW COUNTER', mname, self.counters[mname])                  
               ## Meters: <metric name>:<value>|m
               elif _type == 'm':
                  self.log('METERs', mname, value)
               else: # unknow type, maybe a c[|@<sample rate>]
                  if _type[0] == 'c':
                     self.nb_data += 1
                     if not '|' in _type:
                        continue
                     srate = _type.split('|')[1].strip()
                     if len(srate) == 0 or srate[0] != '@':
                        continue
                     try:
                        rate = float(srate[1:])
                     except ValueError:
                        continue
                     # Invalid rate, 0.0 is invalid too ;)
                     if rate <= 0.0 or rate > 1.0:
                        continue
                     self.log('COUNTER', mname, value, "rate", rate)
                     with self.stats_lock:
                        cvalue, ccount = self.counters.get(mname, (0,0))
                        self.log('INCR counter', (value/rate))
                        self.counters[mname] = (cvalue + (value/rate), ccount + 1/rate)
                        self.log('NEW COUNTER', mname, self.counters[mname])
                     
   
   # Thread for listening to the graphite port in UDP (2003)
   def launch_graphite_udp_listener(self):
      self.graphite_udp_sock = socket.socket(socket.AF_INET, socket.SOCK_DGRAM) # UDP
      self.graphite_udp_sock.setsockopt(socket.SOL_SOCKET, socket.SO_REUSEADDR, 1)
      self.graphite_udp_sock.setsockopt(socket.SOL_SOCKET, socket.SO_RCVBUF, 1048576)
      self.log(self.graphite_udp_sock.getsockopt(socket.SOL_SOCKET, socket.SO_RCVBUF))
      self.graphite_udp_sock.bind((self.addr, self.graphite_port))
      self.log("TS Graphite UDP port open", self.graphite_port)
      self.log("UDP RCVBUF", self.graphite_udp_sock.getsockopt(socket.SOL_SOCKET, socket.SO_RCVBUF))
      while True:#not self.interrupted:
         try:
            data, addr = self.graphite_udp_sock.recvfrom(65535)
         except socket.timeout: # loop until we got some data
            continue
         self.log("UDP Graphite: received message:", len(data), addr)
         STATS.incr('ts.graphite.udp.receive', 1)
         self.graphite_queue.append(data)


   # Same but for the TCP connections
   # TODO: use a real daemon part for this, this is not ok for fast receive
   def launch_graphite_tcp_listener(self):
      self.graphite_tcp_sock = socket.socket(socket.AF_INET, socket.SOCK_STREAM)
      self.graphite_tcp_sock.setsockopt(socket.SOL_SOCKET, socket.SO_REUSEADDR, 1)
      self.graphite_tcp_sock.setsockopt(socket.SOL_SOCKET, socket.SO_RCVBUF, 1048576)
      self.graphite_tcp_sock.bind((self.addr, self.graphite_port))
      self.graphite_tcp_sock.listen(5)
      self.log("TS Graphite TCP port open", self.graphite_port)
      while True:
         try:
            conn, addr = self.graphite_tcp_sock.accept()
         except socket.timeout: # loop until we got some connect
            continue
         conn.settimeout(5.0)
         logger.debug('TCP Graphite Connection address:', addr)
         data = ''
         while 1:
            try:
               ldata = conn.recv(1024)
            except Exception, exp:
               print "TIMEOUT", exp
               break
            if not ldata:
               break
            # Look at only full lines, and not the last part
            # So we look at the position of the last \n
            lst_nidx = ldata.rfind('\n')
            # take all finished lines
            data += ldata[:lst_nidx+1]
            STATS.incr('ts.graphite.tcp.receive', 1)
            self.graphite_queue.append(data)
            # stack the data with the garbage so we will continue it
            # on the next turn
            data = ldata[lst_nidx+1:]
         conn.close()
         # Also stack what the last send
         self.graphite_queue.append(data)


   # Main graphite reaper thread, that will get data from both tcp and udp flow
   # and dispatch it to the others daemons if need
   def graphite_reaper(self):
      while True:
         graphite_queue = self.graphite_queue
         self.graphite_queue = []
         if len(graphite_queue) > 0:
            logger.info("Graphite queue", len(graphite_queue))
         for data in graphite_queue:
            T0 = time.time()
            self.grok_graphite_data(data)
            STATS.timer('ts.graphite.grok-graphite-data', (time.time() - T0)*1000)
         time.sleep(0.1)
   
      
   # Lookup at the graphite lines compat,  run in the graphite-reaper thread
   def grok_graphite_data(self, data):
      STATS.incr('ts.graphite.grok.data', 1)
      forwards = {}
      for line in data.splitlines():
         elts = line.split(' ')
         elts = [s.strip() for s in elts if s.strip()]

         if len(elts) != 3:
            return
         mname, value, timestamp = elts[0], elts[1], elts[2]
         hkey = hashlib.sha1(mname).hexdigest()
         ts_node_manager = self.clust.find_ts_node(hkey)
         # if it's me that manage this key, I add it in my backend
         if ts_node_manager == self.clust.uuid:
             logger.debug("I am the TS node manager")
             try:
                 timestamp = int(timestamp)
             except ValueError:
                 return
             value = self.to_best_int_float(value)
             if value is None:
                 continue
             self.tsb.add_value(timestamp, mname, value)
         # not me? stack a forwarder
         else:
             logger.debug("The node manager for this Ts is ", ts_node_manager)
             l = forwards.get(ts_node_manager, [])
             l.append(line)
             forwards[ts_node_manager] = l

      for (uuid, lst) in forwards.iteritems():
          node = self.clust.nodes.get(uuid, None)
          # maybe the node disapear? bail out, we are not lucky
          if node is None:
              continue
          packets = []
          # first compute the packets
          buf = ''
          for line in lst:
              buf += line+'\n'
              if len(buf) > 1024:
                  packets.append(buf)
                  buf = ''
          if buf != '':
              packets.append(buf)

          # UDP
          sock = socket.socket(socket.AF_INET, socket.SOCK_DGRAM)
          for packet in packets:
             # do NOT use the node['port'], it's the internal communication, not the graphite one!
             sock.sendto(packet, (node['addr'], self.graphite_port))
          sock.close()

          '''
          # TCP mode
          sock = socket.socket(socket.AF_INET, socket.SOCK_STREAM)
          sock.connect( (node['addr'], self.graphite_port) )          
          for packet in packets:
             sock.sendall(packet)
          sock.close()
          '''



import requests
import sys

from .adapters import UnixAdapter

__all__ = ['monkeypatch', 'Session']

DEFAULT_SCHEME = 'http+unix://'


class Session(requests.Session):
    def __init__(self, url_scheme=DEFAULT_SCHEME, *args, **kwargs):
        super(Session, self).__init__(*args, **kwargs)
        self.mount(url_scheme, UnixAdapter())


class monkeypatch(object):
    def __init__(self, url_scheme=DEFAULT_SCHEME):
        self.session = Session()
        requests = self._get_global_requests_module()
        self.orig_requests_get = requests.get
        requests.get = self.session.get

    def _get_global_requests_module(self):
        return sys.modules['requests']

    def __enter__(self):
        return self

    def __exit__(self, *args):
        requests = self._get_global_requests_module()
        requests.get = self.orig_requests_get




import os
import copy
import traceback
import subprocess
try:
    import jinja2
except ImportError:
    jinja2 = None

'''
f = open('/root/kunai/etc/templates/haproxy.cfg')
buf = f.read()
f.close()

t = jinja2.Template(buf)

s = t.render(service=service, nodes=nodes, node=node)
'''

from kunai.log import logger


gclust = None
# Get all nodes that are defining a service sname and where the service is OK
def ok_nodes(service=''):
    global gclust
    sname = service
    with gclust.gossip.nodes_lock:
        nodes = copy.copy(gclust.nodes) # just copy the dict, not the nodes themselves
    res = []
    for n in nodes.values():
        if n['state'] != 'alive':
            continue
        for s in n['services'].values():
            if s['name'] == sname and s['state_id'] == 0:
                res.append( (n,s) )
    return res


class Generator(object):
    def __init__(self, g):
        self.g = g
        print "LAUNCHING"*10, g
        self.buf = None
        self.template = None
        self.output = None

        
    # Open the template file and generate the output
    def generate(self, clust):
        # HACK, atomise this global thing!!
        global gclust
        gclust = clust
        # If not jinja2, bailing out
        if jinja2 is None:
            return
        try:
            f = open(self.g['template'], 'r')
            self.buf = f.read()
            f.close()
        except IOError, exp:
            logger.error('Cannot open template file %s : %s' % (self.g['template'], exp))
            self.buf = None
            self.template = None

        # copy objects because they can move
        node = copy.copy(clust.nodes[clust.uuid])
        with clust.gossip.nodes_lock:
            nodes = copy.copy(clust.nodes) # just copy the dict, not the nodes themselves
            
        # Now try to make it a jinja template object
        try:
            env = jinja2.Environment(trim_blocks=True, lstrip_blocks=True, keep_trailing_newline=True, )
            self.template = env.from_string(self.buf)
        except Exception, exp:
            logger.error('Template file %s did raise an error with jinja2 : %s' % (self.g['template'], exp))
            self.buf = None
            self.template = None

        # Now try to render all of this with real objects
        try:
            self.output = self.template.render(nodes=nodes, node=node, ok_nodes=ok_nodes)
        except Exception, exp:
            logger.error('Template rendering %s did raise an error with jinja2 : %s' % (self.g['template'], traceback.format_exc()))
            self.output = None
            self.template = None
            self.buf = None

            
    # If we did generate, try to see if we must write the file, and if so we will have to
    # launch the command
    def write_if_need(self):
        # Maybe there is no output, if so bail out now :)
        if self.output is None:
            return False

        self.cur_value = ''
        # first try to load the current file if exist and compare to the generated file
        if os.path.exists(self.g['path']):
            try:
                f = open(self.g['path'], 'r')
                self.cur_value = f.read()
                f.close()
            except IOError, exp:
                logger.error('Cannot open path file %s : %s' % (self.g['path'], exp))
                self.output = None
                self.template = ''
                self.buf = ''
                return False
        # If not exists or the value did change, regenerate it :)
        if not os.path.exists(self.g['path']) or self.cur_value != self.output:
            logger.debug('Generator %s generate a new value, writing it to %s' % (self.g['name'], self.g['path']))
            try:
                f = open(self.g['path'], 'w')
                f.write(self.output)
                f.close()
                logger.log('Generator %s did generate a new file at %s' % (self.g['name'], self.g['path']))
                return True
            except IOError, exp:
                logger.error('Cannot write path file %s : %s' % (self.g['path'], exp))
                self.output = None
                self.template = ''
                self.buf = ''
                return False

    # If need launch the restart command, shoul not block too long of
    # course
    def launch_command(self):
        cmd = self.g['command']
        try:
            p = subprocess.Popen(cmd, shell=True, stdout=subprocess.PIPE, stderr=subprocess.PIPE, close_fds=True, preexec_fn=os.setsid)
        except Exception, exp:
            logger.error('Generator %s command launch (%s) fail : %s' % (self.g['name'], cmd, exp))
        output, err = p.communicate()
        rc = p.returncode
        if rc != 0:
            logger.error('Generator %s command launch (%s) error (rc=%s): %s' % (self.g['name'], cmd, rc, '\n'.join([output, err])))
            return
        logger.debug("Generator %s command succeded" % self.g['name'])

        
            



from kunai.log import logger


# Starting a pub-sub system for global set
class PubSub(object):
    def __init__(self):
        self.registered = {}


    # Subscribe to a thread
    def sub(self, k, f):
        if not k in self.registered:
            self.registered[k] = []
        self.registered[k].append(f)

    
    # Call all subscribed function of a thread, and accept args if need
    def pub(self, k, **args):
        if not k in self.registered:
            logger.warning('Call on a pub on a unregistered key %s' % k)
            return
        l = self.registered[k]
        for f in l:
            f(**args)
        

pubsub = PubSub()



try:
    from Crypto.Cipher import AES
    from Crypto.PublicKey import RSA
except ImportError:
    AES = None
    RSA = None


from kunai.log import logger    

    
class Encrypter(object):
    def __init__(self):
        self.encryption_key = None

    def load(self, encryption_key):
        self.encryption_key = encryption_key


    # We received data from UDP, if we are set to encrypt, decrypt it
    def decrypt(self, data):
        if not self.encryption_key:
            return data
        logger.debug('DECRYPT with '+self.encryption_key)
        # Be sure the data is x16 lenght
        if len(data) % 16 != 0:
            data += ' ' * (-len(data) % 16)
        try:
            cyph = AES.new(self.encryption_key, AES.MODE_ECB)
            ndata = cyph.decrypt(data).strip()
            return ndata
        except Exception, exp:
            logger.error('Decryption fail:', exp, part='gossip')
            return ''

        
    def encrypt(self, data):
        if not self.encryption_key:
            return data
        logger.debug('ENCRYPT with '+self.encryption_key)
        # Be sure the data is x16 lenght
        if len(data) % 16 != 0:
            data += ' ' * (-len(data) % 16)
        try:
            cyph = AES.new(self.encryption_key, AES.MODE_ECB)
            ndata = cyph.encrypt(data)
            return ndata
        except Exception, exp:
            logger.error('Encryption fail:', exp, part='gossip')
            return ''
        
    
encrypter = Encrypter()



#!/usr/bin/env python

# Copyright (C) 2014:
#    Gabes Jean, naparuba@gmail.com


import os
import sys
import base64
import uuid
import socket
import time
import json
import requests as rq

from kunai.cluster import Cluster
from kunai.log import cprint, logger
from kunai.version import VERSION
from kunai.launcher import Launcher

# for local unix communication
from kunai.requests_unixsocket import monkeypatch


# Will be populated by the shinken CLI command
CONFIG = None



############# ********************        MEMBERS management          ****************###########

def get_local_socket():
    return CONFIG.get('socket', '/var/lib/kunai/kunai.sock').replace('/', '%2F')

# Get on the local socket. Beware to monkeypatch the get
def get_local(u):
    with monkeypatch():
        p = get_local_socket()
        uri = 'http+unix://%s%s' % (p, u)
        r = rq.get(uri)
        return r


# get a json on the local server, and parse the result    
def get_json(uri):
    try:
        r = get_local(uri)
    except rq.exceptions.ConnectionError, exp:
        logger.error('Cannot connect to local kunai daemon: %s' % exp)
        sys.exit(1)
    try:
        d = json.loads(r.text)
    except ValueError, exp:# bad json
        logger.error('Bad return from the server %s' % exp)
        sys.exit(1)        
    return d
    
    
    

def do_members():
    try:
        r = get_local('/agent/members')
    except rq.exceptions.ConnectionError, exp:
        logger.error(exp)
        return
    try:
        members = json.loads(r.text).values()
    except ValueError, exp:# bad json
        logger.error('Bad return from the server %s' % exp)
        return
    members = sorted(members, key=lambda e:e['name'])
    max_name_size = max([ len(m['name']) for m in members ])
    max_addr_size = max([ len(m['addr']) + len(str(m['port'])) + 1 for m in members ])    
    for m in members:
        name = m['name']
        tags = m['tags']
        port = m['port']
        addr = m['addr']
        state = m['state']
        cprint('%s  ' % name.ljust(max_name_size), end='')
        c = {'alive':'green', 'dead':'red', 'suspect':'yellow', 'leave':'cyan'}.get(state, 'cyan')
        cprint(state, color=c, end='')
        s = ' %s:%s ' % (addr, port)
        s = s.ljust(max_addr_size+2) # +2 for the spaces
        cprint(s, end='')
        cprint(' %s ' % ','.join(tags))        



def do_leave(name=''):
    # Lookup at the localhost name first
    if not name:
        try:
            r = get_local('/agent/name')
        except rq.exceptions.ConnectionError, exp:
            logger.error(exp)
            return
        name = r.text
    try:
        r = get_local('/agent/leave/%s' % name)        
    except rq.exceptions.ConnectionError, exp:
        logger.error(exp)
        return
    
    if r.status_code != 200:
        logger.error('Node %s is missing' % name)
        print r.text
        return
    cprint('Node %s is set to leave state' % name,end='')
    cprint(': OK', color='green')


def do_state(name=''):
    uri = '/agent/state/%s' % name
    if not name:
        uri = '/agent/state'
    try:
        r = get_local(uri)
    except rq.exceptions.ConnectionError, exp:
        logger.error(exp)
        return

    try:
        d = json.loads(r.text)
    except ValueError, exp:# bad json
        logger.error('Bad return from the server %s' % exp)
        return

    print 'Services:'
    for (sname, service) in d['services'].iteritems():
        state = service['state_id']
        cprint('\t%s ' % sname.ljust(20),end='')
        c = {0:'green', 2:'red', 1:'yellow', 3:'cyan'}.get(state, 'cyan')
        state = {0:'OK', 2:'CRITICAL', 1:'WARNING', 3:'UNKNOWN'}.get(state, 'UNKNOWN')
        cprint('%s - ' % state.ljust(8), color=c, end='')
        output = service['check']['output']
        cprint(output.strip())

    print "Checks:"
    for (cname, check) in d['checks'].iteritems():
        state = check['state_id']
        cprint('\t%s ' % cname.ljust(20),end='')
        c = {0:'green', 2:'red', 1:'yellow', 3:'cyan'}.get(state, 'cyan')
        state = {0:'OK', 2:'CRITICAL', 1:'WARNING', 3:'UNKNOWN'}.get(state, 'UNKNOWN')
        cprint('%s - ' % state.ljust(8), color=c, end='')
        output = check['output']
        cprint(output.strip())
        


def do_version():
    cprint(VERSION)

def print_info_title(title):
    #t = title.ljust(15)
    #s = '=================== %s ' % t
    #s += '='*(50 - len(s))
    #cprint(s)
    print '========== [%s]:' % title


def print_2tab(e):
    col_size = 20
    for (k, v) in e:
        s = '%s: ' % k.capitalize()
        s = s.ljust(col_size)
        cprint(s, end='')
        cprint(v, color='green')
    
    
def do_info():
    d = get_json('/agent/info')
    
    logs = d.get('logs')
    pid = d.get('pid')
    name = d.get('name')
    port = d.get('port')
    nb_threads = d.get('threads')['nb_threads']
    httpservers = d.get('httpservers', {'internal':None, 'external':None})
    socket_path = d.get('socket')
    _uuid = d.get('uuid')
    graphite = d.get('graphite')
    statsd = d.get('statsd')
    websocket = d.get('websocket')
    dns = d.get('dns')

    e = [('name', name), ('uuid',_uuid), ('pid', pid), ('port',port), ('socket',socket_path), ('threads', nb_threads)]

    # Normal agent information
    print_info_title('Kunai Daemon')
    print_2tab(e)

    # Normal agent information
    int_server = httpservers['external']
    if int_server:
        e = (('threads', int_server['nb_threads']), ('idle_threads', int_server['idle_threads']), ('queue', int_server['queue']) )
        print_info_title('HTTP (LAN)')
        print_2tab(e)

    # Unix socket http daemon
    int_server = httpservers['internal']
    if int_server:
        e = (('threads', int_server['nb_threads']), ('idle_threads', int_server['idle_threads']), ('queue', int_server['queue']) )
        print_info_title('HTTP (Unix Socket)')
        print_2tab(e)
        
    # Now DNS part
    print_info_title('DNS')
    if dns is None:
        cprint('No dns configured')
    else:
        w = dns
        e = [('enabled', w['enabled']), ('port', w['port']), ('domain',w['domain']) ]
        print_2tab(e)
    
    # Now websocket part
    print_info_title('Websocket')
    if websocket is None:
        cprint('No websocket configured')
    else:
        w = websocket
        st = d.get('websocket_info', None)
        e = [('enabled', w['enabled']), ('port', w['port']) ]
        if st:
            e.append( ('Nb connexions', st.get('nb_connexions')) )
        print_2tab(e)

    # Now graphite part
    print_info_title('Graphite')
    if graphite is None:
        cprint('No graphite configured')
    else:
        g = graphite
        e = [('enabled', g['enabled']), ('port', g['port']), ('udp', g['udp']), ('tcp', g['tcp']) ]
        print_2tab(e)

    # Now statsd part
    print_info_title('Statsd')
    if statsd is None:
        cprint('No statsd configured')
    else:
        s = statsd
        e = [('enabled', s['enabled']), ('port', s['port']), ('interval', s['interval'])]
        print_2tab(e)

    print_info_title('Logs')
    errors  = logs.get('ERROR')
    warnings = logs.get('WARNING')
    e = [ ('errors',len(errors)), ('warnings',len(warnings))]
    print_2tab(e)

    if len(errors) > 0:
        print_info_title('Error logs')
        for s in errors:
            cprint(s, color='red')
    
    if len(warnings) > 0:
        print_info_title('Warning logs')
        for s in warnings:
            cprint(s, color='yellow')
        
    logger.debug('Raw information: %s' % d)
    
    
    
    
# Main daemon function. Currently in blocking mode only
def do_start(daemon):
    cprint('Starting kunai daemon', color='green')
    lock_path = CONFIG.get('lock', '/var/run/kunai.pid')
    l = Launcher(lock_path=lock_path)
    l.do_daemon_init_and_start(is_daemon=daemon)
    # Here only the last son reach this
    l.main()
    
    

def do_stop():
    try:
        r = get_local('/stop')
    except rq.exceptions.ConnectionError, exp:
        logger.error(exp)
        return
    cprint(r.text, color='green')
    
    
    
def do_join(seed=''):
    if seed == '':
        logger.error('Missing target argument. For example 192.168.0.1:6768')
        return
    try:
        r = get_local('/agent/join/%s' % seed)
    except rq.exceptions.ConnectionError, exp:
        logger.error(exp)
        return
    try:
        b = json.loads(r.text)
    except ValueError, exp:# bad json
        logger.error('Bad return from the server %s' % exp)
        return
    cprint('Joining %s : ' % seed, end='')
    if b:
        cprint('OK', color='green')
    else:
        cprint('FAILED', color='red')



def do_keygen():
    k = uuid.uuid1().hex[:16]
    cprint('UDP Encryption key: (aka encryption_key)', end='')
    cprint(base64.b64encode(k), color='green')
    print ''
    try:
        from Crypto.PublicKey import RSA
    except ImportError:
        logger.error('Missing python-crypto module for RSA keys generation, please install it')
        return
    key = RSA.generate(2048)
    privkey = key.exportKey()
    pub_key = key.publickey()
    pubkey = pub_key.exportKey()
    print "Private RSA key (2048). (aka master_key_priv for for file mfkey.priv)"
    cprint(privkey, color='green')
    print ''
    print "Public RSA key (2048). (aka master_key_pub for file mfkey.pub)"
    cprint(pubkey, color='green')
    print ''



def do_exec(tag='*', cmd='uname -a'):
    if cmd == '':
        logger.error('Missing command')
        return
    try:
        r = get_local('/exec/%s?cmd=%s' % (tag, cmd))
    except rq.exceptions.ConnectionError, exp:
        logger.error(exp)
        return
    print r
    cid = r.text
    print "Command group launch as cid", cid
    time.sleep(5) # TODO: manage a real way to get the result..
    try:
        r = get_local('/exec-get/%s' % cid)
    except rq.exceptions.ConnectionError, exp:
        logger.error(exp)
        return
    j = json.loads(r.text)
    #print j
    res = j['res']
    for (uuid, e) in res.iteritems():
        node = e['node']
        nname = node['name']
        color = {'alive':'green', 'dead':'red', 'suspect':'yellow', 'leave':'cyan'}.get(node['state'], 'cyan')
        cprint(nname, color=color)
        cprint('Return code:', end='')
        color = {0:'green', 1:'yellow', 2:'red'}.get(e['rc'], 'cyan')
        cprint(e['rc'], color=color)
        cprint('Output:', end='')
        cprint(e['output'].strip(), color=color)
        if e['err']:
            cprint('Error:', end='')
            cprint(e['err'].strip(), color='red')
        print ''
            

exports = {
    do_members : {
        'keywords': ['members'],
        'args': [],
        'description': 'List the cluster members'
        },

    do_start : {
        'keywords': ['start'],
        'args': [
            {'name' : '--daemon', 'type':'bool', 'default':False, 'description':'Start kunai into the background'},
        ],
        'description': 'Start the kunai daemon'
        },

    do_stop : {
        'keywords': ['stop'],
        'args': [],
        'description': 'Stop the kunai daemon'
        },

    do_version : {
        'keywords': ['version'],
        'args': [],
        'description': 'Print the daemon version'
        },

    do_info : {
        'keywords': ['info'],
        'args': [],
        'description': 'Show info af a daemon'
        },

    do_keygen : {
        'keywords': ['keygen'],
        'args': [],
        'description': 'Generate a encryption key'
        },

    do_exec : {
        'keywords': ['exec'],
        'args': [
            {'name' : 'tag', 'default':'', 'description':'Name of the node tag to execute command on'},
            {'name' : 'cmd', 'default':'uname -a', 'description':'Command to run on the nodes'},
            ],
        'description': 'Execute a command (default to uname -a) on a group of node of the good tag (default to all)'
        },

    do_join : {
        'keywords': ['join'],
        'description': 'Join another node cluster',
        'args': [
            {'name' : 'seed', 'default':'', 'description':'Other node to join. For example 192.168.0.1:6768'},
            ],
        },

    do_leave : {
        'keywords': ['leave'],
        'description': 'Join another node cluster',
        'args': [
            {'name' : 'name', 'default':'', 'description':'Name of the node to force leave. If void, leave our local node'},
            ],
        },


    do_state : {
        'keywords': ['state'],
        'description': 'Print the state of a node',
        'args': [
            {'name' : 'name', 'default':'', 'description':'Name of the node to print state. If void, take our localhost one'},
            ],
        },


    }





# Will manage the global websocket server if need
class WebsocketManager(object):
    def __init__(self):
        self.webso = None

        
    def set(self, webso):
        self.webso = webso
        
        
    def forward(self, msg):
        if self.webso:
            self.webso.send_all(msg)

websocketmgr = WebsocketManager()



#The MIT License (MIT)
#
#Copyright (c) 2013 Dave P.
#
#Permission is hereby granted, free of charge, to any person obtaining a copy
#of this software and associated documentation files (the "Software"), to deal
#in the Software without restriction, including without limitation the rights
#to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
#copies of the Software, and to permit persons to whom the Software is
#furnished to do so, subject to the following conditions:
#
#The above copyright notice and this permission notice shall be included in all
#copies or substantial portions of the Software.
#
#THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
#IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
#FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
#AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
#LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
#OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
#SOFTWARE.

import SocketServer
import hashlib
import base64
import socket
import struct
import ssl
import time
import sys
import errno
import logging
from BaseHTTPServer import BaseHTTPRequestHandler
from StringIO import StringIO
from select import select


class HTTPRequest(BaseHTTPRequestHandler):
	def __init__(self, request_text):
		self.rfile = StringIO(request_text)
		self.raw_requestline = self.rfile.readline()
		self.error_code = self.error_message = None
		self.parse_request()
		

class WebSocket(object):

	handshakeStr = (
		"HTTP/1.1 101 Switching Protocols\r\n"
		"Upgrade: WebSocket\r\n"
		"Connection: Upgrade\r\n"
		"Sec-WebSocket-Accept: %(acceptstr)s\r\n\r\n"
	)
	
	hixiehandshakedStr = (
		"HTTP/1.1 101 WebSocket Protocol Handshake\r\n"
		"Upgrade: WebSocket\r\n"
		"Connection: Upgrade\r\n"
		"Sec-WebSocket-Origin: %(origin)s\r\n"
		"Sec-WebSocket-Location: %(type)s://%(host)s%(location)s\r\n\r\n"
	)
	
	GUIDStr = '258EAFA5-E914-47DA-95CA-C5AB0DC85B11'
	
	STREAM = 0x0
	TEXT = 0x1
	BINARY = 0x2
	CLOSE = 0x8
	PING = 0x9
	PONG = 0xA

	HEADERB1 = 1
	HEADERB2 = 3
	LENGTHSHORT = 4
	LENGTHLONG = 5
	MASK = 6
	PAYLOAD = 7

	def __init__(self, server, sock, address):
		self.server = server
		self.client = sock
		self.address = address
		
		self.handshaked = False
		self.headerbuffer = ''
		self.readdraftkey = False
		self.draftkey = ''
		self.headertoread = 2048
		self.hixie76 = False
		
		self.fin = 0
		self.data = None
		self.opcode = 0
		self.hasmask = 0
		self.maskarray = None
		self.length = 0
		self.lengtharray = None
		self.index = 0
		self.request = None
		self.usingssl = False

		self.state = self.HEADERB1
	
		# restrict the size of header and payload for security reasons
		self.maxheader = 65536
		self.maxpayload = 4194304

	def close(self):
		self.client.close()
		self.state = self.HEADERB1
		self.hasmask = False
		self.handshaked = False
		self.readdraftkey = False
		self.hixie76 = False
		self.headertoread = 2048 
		self.headerbuffer = ''
		self.data = ''


	def handleMessage(self):
		pass

	def handleConnected(self):
		pass

	def handleClose(self):
		pass

	def handlePacket(self):
		# close
		if self.opcode == self.CLOSE:
			self.sendClose()
			raise Exception("received client close")
		# ping
		elif self.opcode == self.PING:
			pass
		
		# pong
		elif self.opcode == self.PONG:
			pass
		
		# data
		elif self.opcode == self.STREAM or self.opcode == self.TEXT or self.opcode == self.BINARY:
			self.handleMessage()	


	def handleData(self):		
		
		# do the HTTP header and handshake
		if self.handshaked is False:
			
			data = self.client.recv(self.headertoread)
			
			if data:
				# accumulate
				self.headerbuffer += data

				if len(self.headerbuffer) >= self.maxheader:
					raise Exception('header exceeded allowable size')

				# we need to read the entire 8 bytes of after the HTTP header, ensure we do
				if self.readdraftkey is True:
					self.draftkey += self.headerbuffer
					read = self.headertoread - len(self.headerbuffer)

					if read != 0:
						self.headertoread = read
					else:
						# complete hixie76 handshake
						self.handshake_hixie76()
					
				# indicates end of HTTP header
				elif '\r\n\r\n' in self.headerbuffer:
					self.request = HTTPRequest(self.headerbuffer)
					# hixie handshake
					if self.request.headers.has_key('Sec-WebSocket-Key1'.lower()) and self.request.headers.has_key('Sec-WebSocket-Key2'.lower()):
						# check if we have the key in our buffer
						index = self.headerbuffer.find('\r\n\r\n') + 4
						# determine how much of the 8 byte key we have
						read = len(self.headerbuffer) - index
						# do we have all the 8 bytes we need?
						if read < 8:
							self.headertoread = 8 - read
							self.readdraftkey = True
							if read > 0:
								self.draftkey += self.headerbuffer[index:index+read]
						
						else:
							# get the key
							self.draftkey += self.headerbuffer[index:index+8]
							# complete hixie handshake
							self.handshake_hixie76()
							
					# handshake rfc 6455
					elif self.request.headers.has_key('Sec-WebSocket-Key'.lower()):
						key = self.request.headers['Sec-WebSocket-Key'.lower()]
						hStr = self.handshakeStr % { 'acceptstr' :  base64.b64encode(hashlib.sha1(key + self.GUIDStr).digest()) }
						self.sendBuffer(hStr)
						self.handshaked = True
						self.headerbuffer = ''
						
						try:
							self.handleConnected()
						except:
							pass		
					else:
						raise Exception('Sec-WebSocket-Key does not exist')

			# remote connection has been closed
			else:
				raise Exception("remote socket closed")
				
		# else do normal data		
		else:
			data = self.client.recv(2048)
			if data:
				for val in data:
					if self.hixie76 is False:
						self.parseMessage(ord(val))
					else:
						self.parseMessage_hixie76(ord(val))
			else:
				raise Exception("remote socket closed")
	


	def handshake_hixie76(self):
	
		k1 = self.request.headers['Sec-WebSocket-Key1'.lower()]
		k2 = self.request.headers['Sec-WebSocket-Key2'.lower()]

		spaces1 = k1.count(" ")
		spaces2 = k2.count(" ")
		num1 = int("".join([c for c in k1 if c.isdigit()])) / spaces1
		num2 = int("".join([c for c in k2 if c.isdigit()])) / spaces2

		key = ''
		key += struct.pack('>I', num1)
		key += struct.pack('>I', num2)
		key += self.draftkey

		typestr = 'ws'
		if self.usingssl is True:
			typestr = 'wss'

		response = self.hixiehandshakedStr % { 'type' : typestr, 'origin' : self.request.headers['Origin'.lower()], 'host' : self.request.headers['Host'.lower()], 'location' : self.request.path }

		self.sendBuffer(response)
		self.sendBuffer(hashlib.md5(key).digest())

		self.handshaked = True
		self.hixie76 = True
		self.headerbuffer = ''

		try:
			self.handleConnected()
		except:
			pass
		

	def sendClose(self):

		msg = bytearray()
		if self.hixie76 is False:
			msg.append(0x88)
			msg.append(0x00)
			self.sendBuffer(msg)
		else:
			pass

	def sendBuffer(self, buff):
		size = len(buff)
		tosend = size
		index = 0

		while tosend > 0:
			try:
				# i should be able to send a bytearray
				sent = self.client.send(str(buff[index:size]))
				if sent == 0:
					raise RuntimeError("socket connection broken")

				index += sent
				tosend -= sent

			except socket.error as e:
				# if we have full buffers then wait for them to drain and try again
				if e.errno == errno.EAGAIN:
					time.sleep(0.001)
				else:
					raise e

	
	#if s is a string then websocket TEXT is sent else BINARY
	def sendMessage(self, s):
		
		if self.hixie76 is False:

			header = bytearray()
			isString = isinstance(s, str)

			if isString is True: 
				header.append(0x81)
			else:
				header.append(0x82)

			b2 = 0		
			length = len(s)

			if length <= 125:
				b2 |= length
				header.append(b2)

			elif length >= 126 and length <= 65535:
				b2 |= 126
				header.append(b2)
				header.extend(struct.pack("!H", length))
		
			else:
				b2 |= 127
				header.append(b2)
				header.extend(struct.pack("!Q", length))

			if length > 0:
				self.sendBuffer(header + s) 
			else:
				self.sendBuffer(header)
			header = None

		else:
			msg = bytearray()			
			msg.append(0)
			if len(s) > 0:
				msg.extend(str(s).encode("UTF8"))
			msg.append(0xFF)

			self.sendBuffer(msg)
			msg = None


	def parseMessage_hixie76(self, byte):

		if self.state == self.HEADERB1:
			if byte == 0:
				self.state = self.PAYLOAD
				self.data = bytearray()

		elif self.state == self.PAYLOAD:
			if byte == 0xFF:
				self.opcode = 1
				self.length = len(self.data)
				try:
					self.handlePacket()
				finally:
					self.data = None
					self.state = self.HEADERB1
			else	:
				self.data.append(byte)
				# if length exceeds allowable size then we except and remove the connection
				if len(self.data) >= self.maxpayload:
					raise Exception('payload exceeded allowable size')


	def parseMessage(self, byte):	
		# read in the header
		if self.state == self.HEADERB1:
			# fin
			self.fin = (byte & 0x80)
			# get opcode
			self.opcode = (byte & 0x0F)
			
			self.state = self.HEADERB2
		
		elif self.state == self.HEADERB2:
			mask = byte & 0x80				
			length = byte & 0x7F
			
			if mask == 128:
				self.hasmask = True
			else:
				self.hasmask = False
			
			if length <= 125:
				self.length = length
				
				# if we have a mask we must read it
				if self.hasmask is True:
					self.maskarray = bytearray()
					self.state = self.MASK
				else:
					# if there is no mask and no payload we are done
					if self.length <= 0:
						try:
							self.handlePacket()
						finally:
							self.state = self.HEADERB1
							self.data = None
							
					# we have no mask and some payload
					else:
						self.index = 0
						self.data = bytearray()
						self.state = self.PAYLOAD
					
			elif length == 126:
				self.lengtharray = bytearray()
				self.state = self.LENGTHSHORT
				
			elif length == 127:
				self.lengtharray = bytearray()
				self.state = self.LENGTHLONG

		
		elif self.state == self.LENGTHSHORT:
			self.lengtharray.append(byte)

			if len(self.lengtharray) > 2:
				raise Exception('short length exceeded allowable size')

			if len(self.lengtharray) == 2:
				self.length = struct.unpack_from('!H', str(self.lengtharray))[0]
				
				if self.hasmask is True:
					self.maskarray = bytearray()
					self.state = self.MASK
				else:
					# if there is no mask and no payload we are done
					if self.length <= 0:
						try:
							self.handlePacket()
						finally:
							self.state = self.HEADERB1
							self.data = None

					# we have no mask and some payload
					else:
						self.index = 0
						self.data = bytearray()
						self.state = self.PAYLOAD
			
		elif self.state == self.LENGTHLONG:

			self.lengtharray.append(byte)

			if len(self.lengtharray) > 8:
				raise Exception('long length exceeded allowable size')

			if len(self.lengtharray) == 8:
				self.length = struct.unpack_from('!Q', str(self.lengtharray))[0]

				if self.hasmask is True:
					self.maskarray = bytearray()
					self.state = self.MASK
				else:
					# if there is no mask and no payload we are done
					if self.length <= 0:
						try:
							self.handlePacket()
						finally:
							self.state = self.HEADERB1
							self.data = None

					# we have no mask and some payload
					else:
						self.index = 0
						self.data = bytearray()
						self.state = self.PAYLOAD
			
		# MASK STATE
		elif self.state == self.MASK:
			self.maskarray.append(byte)

			if len(self.maskarray) > 4:
				raise Exception('mask exceeded allowable size')

			if len(self.maskarray) == 4:
				# if there is no mask and no payload we are done
				if self.length <= 0:
					try:
						self.handlePacket()
					finally:
						self.state = self.HEADERB1
						self.data = None
						
				# we have no mask and some payload
				else:
					self.index = 0
					self.data = bytearray()
					self.state = self.PAYLOAD
		
		# PAYLOAD STATE
		elif self.state == self.PAYLOAD:
			if self.hasmask is True:
				self.data.append( byte ^ self.maskarray[self.index % 4] )
			else:
				self.data.append( byte )

			# if length exceeds allowable size then we except and remove the connection
			if len(self.data) >= self.maxpayload:
				raise Exception('payload exceeded allowable size')

			# check if we have processed length bytes; if so we are done
			if (self.index+1) == self.length:
				try:
					self.handlePacket()
				finally:
					self.state = self.HEADERB1
					self.data = None
			else:
				self.index += 1


class SimpleWebSocketServer(object):
	def __init__(self, host, port, websocketclass):
		self.websocketclass = websocketclass
		self.serversocket = socket.socket(socket.AF_INET, socket.SOCK_STREAM)
		self.serversocket.setsockopt(socket.SOL_SOCKET, socket.SO_REUSEADDR, 1)
		self.serversocket.bind((host, port))
		self.serversocket.listen(5)
		self.connections = {}
		self.listeners = [self.serversocket]


	def decorateSocket(self, sock):
		return sock

	def constructWebSocket(self, sock, address):
		return self.websocketclass(self, sock, address)

	def close(self):
		self.serversocket.close()
	
		for conn in self.connections.itervalues():
			try:
				conn.handleClose()
			except:
				pass
	
			conn.close()


	def serveforever(self):
		while True:
			rList, wList, xList = select(self.listeners, [], self.listeners, 1)	

			for ready in rList:
				if ready == self.serversocket:
					try:
						sock, address = self.serversocket.accept()
						newsock = self.decorateSocket(sock)
						newsock.setblocking(0)
						fileno = newsock.fileno()
						self.listeners.append(fileno)
						self.connections[fileno] = self.constructWebSocket(newsock, address)

					except Exception as n:

						logging.debug(str(address) + ' ' + str(n))

						if sock is not None:
							sock.close()
				else:
					client = self.connections[ready]
					fileno = client.client.fileno()
				
					try:
						client.handleData()

					except Exception as n:

						logging.debug(str(client.address) + ' ' + str(n))

						try:
							client.handleClose()
						except:
							pass

						client.close()

						del self.connections[fileno]
						self.listeners.remove(ready)
		
			for failed in xList:
				if failed == self.serversocket:
					self.close()
					raise Exception("server socket failed")
				else:
					client = self.connections[failed]
					fileno = client.client.fileno()

					try:
						client.handleClose()
					except:
						pass

					client.close()

					del self.connections[fileno]
					self.listeners.remove(failed)
					

class SimpleSSLWebSocketServer(SimpleWebSocketServer):

	def __init__(self, host, port, websocketclass, certfile, keyfile, version = ssl.PROTOCOL_TLSv1):

		SimpleWebSocketServer.__init__(self, host, port, websocketclass)

		self.cerfile = certfile
		self.keyfile = keyfile
		self.version = version

	def close(self):
		super(SimpleSSLWebSocketServer, self).close()

	def decorateSocket(self, sock):
		sslsock = ssl.wrap_socket(sock,
		                     server_side=True,
		                     certfile=self.cerfile,
		                     keyfile=self.keyfile,
		                     ssl_version=self.version)
		return sslsock

	def constructWebSocket(self, sock, address):
		ws = self.websocketclass(self, sock, address)
		ws.usingssl = True
		return ws

	def serveforever(self):
		super(SimpleSSLWebSocketServer, self).serveforever()
					



import platform
import traceback
import subprocess
from collections import defaultdict

from kunai.stats import STATS
from kunai.log import logger
from kunai.threadmgr import threader
from kunai.now import NOW


pythonVersion = platform.python_version_tuple()
python24 = platform.python_version().startswith('2.4')


class Collector:
    class __metaclass__(type):
        __inheritors__ = set()
        def __new__(meta, name, bases, dct):
            klass = type.__new__(meta, name, bases, dct)
            meta.__inheritors__.add(klass)
            return klass

    @classmethod
    def get_sub_class(cls):        
        return cls.__inheritors__

    
    def __init__(self, config):
        self.config = config
        self.pythonVersion = pythonVersion
        
        self.mysqlConnectionsStore = None
        self.mysqlSlowQueriesStore = None
        self.mysqlVersion = None
        self.networkTrafficStore = {}
        self.nginxRequestsStore = None
        self.mongoDBStore = None
        self.apacheTotalAccesses = None
        self.plugins = None
        self.topIndex = 0
        self.os = None
        self.linuxProcFsLocation = None


    # Execute a shell command and return the result or '' if there is an error
    def execute_shell(self, cmd):
        # Get output from a command
        logger.debug('execute_shell:: %s' % cmd)
        output = ''
        try:
            try:
                proc = subprocess.Popen(cmd.split(' '), stdout=subprocess.PIPE, close_fds=True)
                output, err = proc.communicate()
                if int(pythonVersion[1]) >= 6:
                    try:
                        proc.kill()
                    except Exception, e:
                        pass
                if err:
                    logger.error('Error in sub process', err)

            except Exception, e:
                logger.error('execute_shell exception = %s', traceback.format_exc())
                return False

        finally:
            if int(pythonVersion[1]) >= 6:
                try:
                    proc.kill()
                except Exception, e:
                    #logger.debug('Process already terminated')
                    pass
            return output


    # from a dict recursivly build a ts
    # 'bla':{'foo':bar, 'titi': toto} => bla.foo.bar bla.titi.toto
    def create_ts_from_data(self, d, l, s):
        #print 'Create ts from data', d
        if not isinstance(d, dict):
            if isinstance(d, basestring):
                #print "REFUSING BAD VALUE", d
                return
            if isinstance(d, float) or isinstance(d, int):
                #print "FINISH HIM!"
                _t = l[:]
                #_t.append(d)
                _nts = '.'.join(_t)#ts, d)
                nts = '%s %s' % (_nts, d)
                s.add(nts)
            return
        # For each key,
        for (k,v) in d.iteritems():
            nl = l[:] # use a copy to l so it won't be overwriten
            nl.append(k)
            self.create_ts_from_data(v, nl, s)
            
        
        

    def main(self):
        logger.debug('Launching main for %s' % self.__class__)
        try:
            r = self.launch()
        except Exception, exp:
            logger.error('Collector %s main error: %s' (self.__class__.__name__.lower(), traceback.format_exc()))
            return
        #logger.debug('COLRESULT', r, self.__class__.__name__.lower())
        s = set()
        self.create_ts_from_data(r, [], s)
        print "TSS=", s
        



import httplib # Used only for handling httplib.HTTPException (case #26701)
import os
import platform
import re
import urllib
import urllib2
import traceback
import time
from StringIO import StringIO


from kunai.log import logger
from kunai.collector import Collector


class LoadAverage(Collector):
    def launch(self):
        logger.debug('getLoadAvrgs: start')

        # Get the triplet from the python function
        try:
            loadAvrgs_1, loadAvrgs_5, loadAvrgs_15 = os.getloadavg()
        except OSError:
            # If not available, return nothing
            return False
        
        logger.debug('getLoadAvrgs: parsing')

        loadAvrgs = {'1': loadAvrgs_1, '5': loadAvrgs_5, '15': loadAvrgs_15}
        
        logger.debug('getLoadAvrgs: completed, returning')

        return loadAvrgs



import httplib # Used only for handling httplib.HTTPException (case #26701)
import os
import platform
import re
import urllib
import urllib2
import traceback
import time
from StringIO import StringIO


from kunai.log import logger
from kunai.collector import Collector


class DiskUsage(Collector):
    def launch(self):
        #logger.debug('getDiskUsage: start')

        #logger.debug('getDiskUsage: attempting Popen')

        _cmd = 'df -k -x smbfs -x tmpfs -x cifs -x iso9660 -x udf -x nfsv4'
        df = self.execute_shell(_cmd)
        if not df:
            return False
                
        #logger.debug('getDiskUsage: Popen success, start parsing')

        # Split out each volume
        volumes = df.split('\n')

        #logger.debug('getDiskUsage: parsing, split')

        # Remove first (headings) and last (blank)
        volumes.pop(0)
        volumes.pop()

        #logger.debug('getDiskUsage: parsing, pop')

        usageData = []

        regexp = re.compile(r'([0-9]+)')

        # Set some defaults
        previousVolume = None
        volumeCount = 0

        #logger.debug('getDiskUsage: parsing, start loop')

        for volume in volumes:
            #logger.debug('getDiskUsage: parsing volume: %s', volume)

            # Split out the string
            volume = volume.split(None, 10)

            # Handle df output wrapping onto multiple lines (case 27078 and case 30997)
            # Thanks to http://github.com/sneeu
            if len(volume) == 1: # If the length is 1 then this just has the mount name
                previousVolume = volume[0] # We store it, then continue the for
                continue

            if previousVolume != None: # If the previousVolume was set (above) during the last loop
                volume.insert(0, previousVolume) # then we need to insert it into the volume
                previousVolume = None # then reset so we don't use it again

            volumeCount = volumeCount + 1

            # Sometimes the first column will have a space, which is usually a system line that isn't relevant
            # e.g. map -hosts              0         0          0   100%    /net
            # so we just get rid of it
            # Also ignores lines with no values
            if re.match(regexp, volume[1]) == None or re.match(regexp, volume[2]) == None or re.match(regexp, volume[3]) == None:
                pass
            else:
                #print "DUMPING VOLUME", volume
                try:
                    volume[1] = int(volume[1]) / 1024  # total
                    volume[2] = int(volume[2]) / 1024  # Used
                    volume[3] = int(volume[3]) / 1024  # Available
                except Exception, e:
                    logger.error('getDiskUsage: parsing, loop %s - Used or Available not present' % (repr(e),))
                usageData.append(volume)

        return usageData




import httplib # Used only for handling httplib.HTTPException (case #26701)
import os
import sys
import platform
import re
import urllib
import urllib2
import traceback
import time
from StringIO import StringIO


from kunai.log import logger
from kunai.collector import Collector


class IoStats(Collector):
    def launch(self):
        #logger.debug('get_open_ports: start')

        open_ports = {'tcp' : [], 'udp' : []}

        if sys.platform != 'linux2':
            logger.debug('get_open_ports: unsupported platform')
            return False

        #logger.debug('get_open_ports: linux2')

        
        try:
            _cmd = 'netstat -tuln'
            netstat = self.execute_shell(_cmd)
            if not netstat:
                logger.error('get_open_ports: exception in launching command')
                return False
            
            for line in netstat.splitlines():
                line = line.strip()

                # Will be something like
                #tcp        0      0 0.0.0.0:27017           0.0.0.0:*               LISTEN
                if not line.startswith('tcp') and not line.startswith('udp'):
                    # Not a good line, skip it
                    continue
                #print "LOOKING AT LINE"
                elts = [ e for e in line.split(' ') if e]
                #print "ELEMENTS", elts
                if len(elts) != 6:
                    #print "BAD LINE", elts
                    continue
                    
                open_port = {}
                open_port['proto']  = elts[0]
                open_port['source'] = elts[3]
                open_port['dest']   = elts[4]
                open_port['state']  = elts[5]

                if open_port['proto'].startswith('tcp'):
                    open_ports['tcp'].append(open_port)
                elif open_port['proto'].startswith('udp'):
                    open_ports['udp'].append(open_port)
                else:
                    print "Unknown protocol??"

        except Exception, ex:
            logger.error('get_open_ports: exception = %s', traceback.format_exc())
            return False

        #logger.debug('get_open_ports: completed, returning')
        return open_ports



import httplib # Used only for handling httplib.HTTPException (case #26701)
import os
import platform
import re
import urllib
import urllib2
import traceback
import time
from StringIO import StringIO


from kunai.log import logger
from kunai.collector import Collector


class Mongodb(Collector):
    def launch(self):
        logger.debug('getMongoDBStatus: start')

        if 'MongoDBServer' not in self.config or self.config['MongoDBServer'] == '':
            logger.debug('getMongoDBStatus: config not set')
            return False

        logger.debug('getMongoDBStatus: config set')

        try:
            import pymongo
            from pymongo import Connection

        except ImportError:
            logger.error('Unable to import pymongo library')
            return False

        # The dictionary to be returned.
        mongodb = {}

        try:
            import urlparse
            parsed = urlparse.urlparse(self.config['MongoDBServer'])

            mongoURI = ''

            # Can't use attributes on Python 2.4
            if parsed[0] != 'mongodb':

                mongoURI = 'mongodb://'

                if parsed[2]:

                    if parsed[0]:

                        mongoURI = mongoURI + parsed[0] + ':' + parsed[2]

                    else:
                        mongoURI = mongoURI + parsed[2]

            else:

                mongoURI = self.config['MongoDBServer']

            logger.debug('-- mongoURI: %s', mongoURI)

            conn = Connection(mongoURI, slave_okay=True)

            logger.debug('Connected to MongoDB')

        except Exception, ex:
            logger.error('Unable to connect to MongoDB server %s - Exception = %s', mongoURI, traceback.format_exc())
            return False

        # Older versions of pymongo did not support the command()
        # method below.
        try:
            db = conn['local']

            # Server status
            statusOutput = db.command('serverStatus') # Shorthand for {'serverStatus': 1}

            logger.debug('getMongoDBStatus: executed serverStatus')

            # Setup
            import datetime
            status = {}

            # Version
            try:
                status['version'] = statusOutput['version']

                logger.debug('getMongoDBStatus: version %s', statusOutput['version'])

            except KeyError, ex:
                logger.error('getMongoDBStatus: version KeyError exception = %s', ex)
                pass

            # Global locks
            try:
                logger.debug('getMongoDBStatus: globalLock')

                status['globalLock'] = {}
                status['globalLock']['ratio'] = statusOutput['globalLock']['ratio']

                status['globalLock']['currentQueue'] = {}
                status['globalLock']['currentQueue']['total'] = statusOutput['globalLock']['currentQueue']['total']
                status['globalLock']['currentQueue']['readers'] = statusOutput['globalLock']['currentQueue']['readers']
                status['globalLock']['currentQueue']['writers'] = statusOutput['globalLock']['currentQueue']['writers']

            except KeyError, ex:
                logger.error('getMongoDBStatus: globalLock KeyError exception = %s', ex)
                pass

            # Memory
            try:
                logger.debug('getMongoDBStatus: memory')

                status['mem'] = {}
                status['mem']['resident'] = statusOutput['mem']['resident']
                status['mem']['virtual'] = statusOutput['mem']['virtual']
                status['mem']['mapped'] = statusOutput['mem']['mapped']

            except KeyError, ex:
                logger.error('getMongoDBStatus: memory KeyError exception = %s', ex)
                pass

            # Connections
            try:
                logger.debug('getMongoDBStatus: connections')

                status['connections'] = {}
                status['connections']['current'] = statusOutput['connections']['current']
                status['connections']['available'] = statusOutput['connections']['available']

            except KeyError, ex:
                logger.error('getMongoDBStatus: connections KeyError exception = %s', ex)
                pass

            # Extra info (Linux only)
            try:
                logger.debug('getMongoDBStatus: extra info')

                status['extraInfo'] = {}
                status['extraInfo']['heapUsage'] = statusOutput['extra_info']['heap_usage_bytes']
                status['extraInfo']['pageFaults'] = statusOutput['extra_info']['page_faults']

            except KeyError, ex:
                logger.debug('getMongoDBStatus: extra info KeyError exception = %s', ex)
                pass

            # Background flushing
            try:
                logger.debug('getMongoDBStatus: backgroundFlushing')

                status['backgroundFlushing'] = {}
                delta = datetime.datetime.utcnow() - statusOutput['backgroundFlushing']['last_finished']
                status['backgroundFlushing']['secondsSinceLastFlush'] = delta.seconds
                status['backgroundFlushing']['lastFlushLength'] = statusOutput['backgroundFlushing']['last_ms']
                status['backgroundFlushing']['flushLengthAvrg'] = statusOutput['backgroundFlushing']['average_ms']

            except KeyError, ex:
                logger.debug('getMongoDBStatus: backgroundFlushing KeyError exception = %s', ex)
                pass

            # Per second metric calculations (opcounts and asserts)
            try:
                if self.mongoDBStore == None:
                    logger.debug('getMongoDBStatus: per second metrics no cached data, so storing for first time')
                    self.setMongoDBStore(statusOutput)

                else:
                    logger.debug('getMongoDBStatus: per second metrics cached data exists')

                    accessesPS = float(statusOutput['indexCounters']['btree']['accesses'] - self.mongoDBStore['indexCounters']['btree']['accessesPS']) / 60

                    if accessesPS >= 0:
                        status['indexCounters'] = {}
                        status['indexCounters']['btree'] = {}
                        status['indexCounters']['btree']['accessesPS'] = accessesPS
                        status['indexCounters']['btree']['hitsPS'] = float(statusOutput['indexCounters']['btree']['hits'] - self.mongoDBStore['indexCounters']['btree']['hitsPS']) / 60
                        status['indexCounters']['btree']['missesPS'] = float(statusOutput['indexCounters']['btree']['misses'] - self.mongoDBStore['indexCounters']['btree']['missesPS']) / 60
                        status['indexCounters']['btree']['missRatioPS'] = float(statusOutput['indexCounters']['btree']['missRatio'] - self.mongoDBStore['indexCounters']['btree']['missRatioPS']) / 60

                        status['opcounters'] = {}
                        status['opcounters']['insertPS'] = float(statusOutput['opcounters']['insert'] - self.mongoDBStore['opcounters']['insertPS']) / 60
                        status['opcounters']['queryPS'] = float(statusOutput['opcounters']['query'] - self.mongoDBStore['opcounters']['queryPS']) / 60
                        status['opcounters']['updatePS'] = float(statusOutput['opcounters']['update'] - self.mongoDBStore['opcounters']['updatePS']) / 60
                        status['opcounters']['deletePS'] = float(statusOutput['opcounters']['delete'] - self.mongoDBStore['opcounters']['deletePS']) / 60
                        status['opcounters']['getmorePS'] = float(statusOutput['opcounters']['getmore'] - self.mongoDBStore['opcounters']['getmorePS']) / 60
                        status['opcounters']['commandPS'] = float(statusOutput['opcounters']['command'] - self.mongoDBStore['opcounters']['commandPS']) / 60

                        status['asserts'] = {}
                        status['asserts']['regularPS'] = float(statusOutput['asserts']['regular'] - self.mongoDBStore['asserts']['regularPS']) / 60
                        status['asserts']['warningPS'] = float(statusOutput['asserts']['warning'] - self.mongoDBStore['asserts']['warningPS']) / 60
                        status['asserts']['msgPS'] = float(statusOutput['asserts']['msg'] - self.mongoDBStore['asserts']['msgPS']) / 60
                        status['asserts']['userPS'] = float(statusOutput['asserts']['user'] - self.mongoDBStore['asserts']['userPS']) / 60
                        status['asserts']['rolloversPS'] = float(statusOutput['asserts']['rollovers'] - self.mongoDBStore['asserts']['rolloversPS']) / 60

                        self.setMongoDBStore(statusOutput)
                    else:
                        logger.debug('getMongoDBStatus: per second metrics negative value calculated, mongod likely restarted, so clearing cache')
                        self.setMongoDBStore(statusOutput)

            except KeyError, ex:
                logger.error('getMongoDBStatus: per second metrics KeyError exception = %s', ex)
                pass

            # Cursors
            try:
                logger.debug('getMongoDBStatus: cursors')

                status['cursors'] = {}
                status['cursors']['totalOpen'] = statusOutput['cursors']['totalOpen']

            except KeyError, ex:
                logger.error('getMongoDBStatus: cursors KeyError exception = %s', ex)
                pass

            # Replica set status
            if 'MongoDBReplSet' in self.config and self.config['MongoDBReplSet'] == 'yes':
                logger.debug('getMongoDBStatus: get replset status too')

                # isMaster (to get state
                isMaster = db.command('isMaster')

                logger.debug('getMongoDBStatus: executed isMaster')

                status['replSet'] = {}
                status['replSet']['setName'] = isMaster['setName']
                status['replSet']['isMaster'] = isMaster['ismaster']
                status['replSet']['isSecondary'] = isMaster['secondary']

                if 'arbiterOnly' in isMaster:
                    status['replSet']['isArbiter'] = isMaster['arbiterOnly']

                logger.debug('getMongoDBStatus: finished isMaster')

                # rs.status()
                db = conn['admin']
                replSet = db.command('replSetGetStatus')

                logger.debug('getMongoDBStatus: executed replSetGetStatus')

                status['replSet']['myState'] = replSet['myState']

                status['replSet']['members'] = {}

                for member in replSet['members']:

                    logger.debug('getMongoDBStatus: replSetGetStatus looping %s', member['name'])

                    status['replSet']['members'][str(member['_id'])] = {}

                    status['replSet']['members'][str(member['_id'])]['name'] = member['name']
                    status['replSet']['members'][str(member['_id'])]['state'] = member['state']

                    # Optime delta (only available from not self)
                    # Calculation is from http://docs.python.org/library/datetime.html#datetime.timedelta.total_seconds
                    if 'optimeDate' in member: # Only available as of 1.7.2
                        deltaOptime = datetime.datetime.utcnow() - member['optimeDate']
                        status['replSet']['members'][str(member['_id'])]['optimeDate'] = (deltaOptime.microseconds + (deltaOptime.seconds + deltaOptime.days * 24 * 3600) * 10**6) / 10**6

                    if 'self' in member:
                        status['replSet']['myId'] = member['_id']

                    # Have to do it manually because total_seconds() is only available as of Python 2.7
                    else:
                        if 'lastHeartbeat' in member:
                            deltaHeartbeat = datetime.datetime.utcnow() - member['lastHeartbeat']
                            status['replSet']['members'][str(member['_id'])]['lastHeartbeat'] = (deltaHeartbeat.microseconds + (deltaHeartbeat.seconds + deltaHeartbeat.days * 24 * 3600) * 10**6) / 10**6

                    if 'errmsg' in member:
                        status['replSet']['members'][str(member['_id'])]['error'] = member['errmsg']

            # db.stats()
            if 'MongoDBDBStats' in self.config and self.config['MongoDBDBStats'] == 'yes':
                logger.debug('getMongoDBStatus: db.stats() too')

                status['dbStats'] = {}

                for database in conn.database_names():

                    if database != 'config' and database != 'local' and database != 'admin' and database != 'test':

                        logger.debug('getMongoDBStatus: executing db.stats() for %s', database)

                        status['dbStats'][database] = conn[database].command('dbstats')
                        status['dbStats'][database]['namespaces'] = conn[database]['system']['namespaces'].count()

                        # Ensure all strings to prevent JSON parse errors. We typecast on the server
                        for key in status['dbStats'][database].keys():
                            status['dbStats'][database][key] = str(status['dbStats'][database][key])


        except Exception, ex:
            logger.error('Unable to get MongoDB status - Exception = %s', traceback.format_exc())
            return False

        logger.debug('getMongoDBStatus: completed, returning')

        return status




    def setMongoDBStore(self, statusOutput):
        self.mongoDBStore = {}

        self.mongoDBStore['indexCounters'] = {}
        self.mongoDBStore['indexCounters']['btree'] = {}
        self.mongoDBStore['indexCounters']['btree']['accessesPS'] = statusOutput['indexCounters']['btree']['accesses']
        self.mongoDBStore['indexCounters']['btree']['hitsPS'] = statusOutput['indexCounters']['btree']['hits']
        self.mongoDBStore['indexCounters']['btree']['missesPS'] = statusOutput['indexCounters']['btree']['misses']
        self.mongoDBStore['indexCounters']['btree']['missRatioPS'] = statusOutput['indexCounters']['btree']['missRatio']

        self.mongoDBStore['opcounters'] = {}
        self.mongoDBStore['opcounters']['insertPS'] = statusOutput['opcounters']['insert']
        self.mongoDBStore['opcounters']['queryPS'] = statusOutput['opcounters']['query']
        self.mongoDBStore['opcounters']['updatePS'] = statusOutput['opcounters']['update']
        self.mongoDBStore['opcounters']['deletePS'] = statusOutput['opcounters']['delete']
        self.mongoDBStore['opcounters']['getmorePS'] = statusOutput['opcounters']['getmore']
        self.mongoDBStore['opcounters']['commandPS'] = statusOutput['opcounters']['command']

        self.mongoDBStore['asserts'] = {}
        self.mongoDBStore['asserts']['regularPS'] = statusOutput['asserts']['regular']
        self.mongoDBStore['asserts']['warningPS'] = statusOutput['asserts']['warning']
        self.mongoDBStore['asserts']['msgPS'] = statusOutput['asserts']['msg']
        self.mongoDBStore['asserts']['userPS'] = statusOutput['asserts']['user']
        self.mongoDBStore['asserts']['rolloversPS'] = statusOutput['asserts']['rollovers']



import threading
import ctypes
import traceback
import cStringIO

from kunai.log import logger
from kunai.pubsub import pubsub

# this part is doomed for windows portability, will be fun to manage :)
try:
    libc = ctypes.CDLL('libc.so.6')
except Exception:
    libc = None



class ThreadMgr(object):
    def __init__(self):
        self.all_threads = []


    def check_alives(self):
        self.all_threads = [t for t in self.all_threads if t.is_alive()]                


    def get_info(self):
        return {'nb_threads' : len(self.all_threads)}
    

    def create_and_launch(self, f, args=(), name='unamed-thread', essential=False):
        def w(is_essential):
            tid = 0
            if libc:
                tid = libc.syscall(186) # get the threadid when you are in it :)
            logger.info('THREAD launch (%s) with thread id (%d)' % (name, tid))
            try:
                f(*args)
            except Exception, exp:
                output = cStringIO.StringIO()
                traceback.print_exc(file=output)
                logger.error("Thread %s is exiting on error. Back trace of this error: %s" % (name, output.getvalue()))
                output.close()

                if is_essential:
                    # Maybe the thread WAS an essential one (like http thread or something like this), if so
                    # catch it and close the whole daemon
                    logger.error('The thread %s was an essential one, we are stopping the daemon do not be in an invalid state' % name)
                    pubsub.pub('interrupt')                

        # Create a daemon thread with our wrapper function that will manage initial logging
        # and exception catchs
        t = threading.Thread(None, target=w, name=name, args=(essential,))
        t.daemon = True
        t.start()
        self.all_threads.append(t)
        return t


threader = ThreadMgr()




import socket


UDP_IP = "192.168.56.104"
UDP_PORT = 8125


### Don't espect this part to work on your server without modifing the IP :)



#gauge = ' moncul.cpu.used : 90 |g'
#counter = ' mabite : 2 |c|@0.1'
#timer = ' montimer : 0.001 | ms'


class Stats(object):
   def __init__(self):
      self.stats = {}
      self.sock = socket.socket(socket.AF_INET, socket.SOCK_DGRAM)
      self.buf = []

   def stack(self, s):
      self.buf.append(s)
      if len(self.buf) > 20:
         self.sock.sendto('\n'.join(self.buf), (UDP_IP, UDP_PORT))
         self.buf = []


   # Will increment a stat key, if None, start at 0
   def incr(self, k, v):
      self.stats[k] = self.stats.get(k, 0) + v
      s = '%s: %d |c' % (k,v)
      self.stack(s)

   # Will increment a stat key, if None, start at 0
   def timer(self, k, v):
      s = '%s: %f |ms' % (k,v)
      self.stack(s)


   def get(self, k):
      return self.stats.get(k, 0)


   def show(self):
      return ', '.join(['%s:%.4f' % (k,v) for (k,v) in self.stats.iteritems()])


STATS = Stats()



import httplib # Used only for handling httplib.HTTPException (case #26701)
import os
import sys
import platform
import re
import urllib
import urllib2
import traceback
import time
from StringIO import StringIO
from multiprocessing import subprocess

from kunai.log import logger
from kunai.collector import Collector


class CpuStats(Collector):
    def launch(self):
        logger.debug('getCPUStats: start')

        cpuStats = {}

        if sys.platform == 'linux2':
            logger.debug('getCPUStats: linux2')

            headerRegexp = re.compile(r'.*?([%][a-zA-Z0-9]+)[\s+]?')
            itemRegexp = re.compile(r'.*?\s+(\d+)[\s+]?')
            valueRegexp = re.compile(r'\d+\.\d+')
            proc = None
            try:
                proc = subprocess.Popen(['mpstat', '-P', 'ALL', '1', '1'], stdout=subprocess.PIPE, close_fds=True)
                stats = proc.communicate()[0]

                if int(self.pythonVersion[1]) >= 6:
                    try:
                        proc.kill()
                    except Exception, e:
                        logger.debug('Process already terminated')

                stats = stats.split('\n')
                header = stats[2]
                headerNames = re.findall(headerRegexp, header)
                device = None

                for statsIndex in range(4, len(stats)): # skip "all"
                    row = stats[statsIndex]

                    if not row: # skip the averages
                        break

                    deviceMatch = re.match(itemRegexp, row)

                    if deviceMatch is not None:
                        device = 'CPU%s' % deviceMatch.groups()[0]

                    values = re.findall(valueRegexp, row.replace(',', '.'))

                    cpuStats[device] = {}
                    for headerIndex in range(0, len(headerNames)):
                        headerName = headerNames[headerIndex]
                        cpuStats[device][headerName] = values[headerIndex]

            except OSError, ex:
                # we dont have it installed return nothing
                return False

            except Exception, ex:
                if int(self.pythonVersion[1]) >= 6:
                    try:
                        if proc:
                            proc.kill()
                    except UnboundLocalError, e:
                        logger.debug('Process already terminated')
                    except Exception, e:
                        logger.debug('Process already terminated')

                logger.error('getCPUStats: exception = %s', traceback.format_exc())
                return False
        else:
            logger.debug('getCPUStats: unsupported platform')
            return False

        logger.debug('getCPUStats: completed, returning')
        return cpuStats




#!/usr/bin/env python
# Copyright (C) 2014:
#    Gabes Jean, naparuba@gmail.com

import copy
import time
from kunai_test import *

from kunai.threadmgr import ThreadMgr

i = 0

def f():
    global i
    print "I am a thread"
    i += 1
    time.sleep(0.5)

class TestThreads(KunaiTest):
    def setUp(self):
        pass

    def test_thread(self):
        T = ThreadMgr()
        T.create_and_launch(f, name='F')
        self.assert_(len(T.all_threads) == 1)
        T.check_alives()
        time.sleep(1)
        T.check_alives()
        self.assert_(len(T.all_threads) == 0)
        self.assertIs(i, 1)


if __name__ == '__main__':
    unittest.main()



#!/usr/bin/env python
# Copyright (C) 2014:
#    Gabes Jean, naparuba@gmail.com

import copy
import time
import threading
from kunai_test import *

from kunai.gossip import Gossip
from kunai.broadcast import Broadcaster


class TestGossip(KunaiTest):
    def setUp(self):
        self.gossip = Gossip({}, threading.RLock(), 'localhost', 6768, 'testing-kunai', 0, 'AAAA', ['linux', 'kv'], [], False)

    def test_gossip(self):
        pass

if __name__ == '__main__':
    unittest.main()



import os
import sys
import base64
import uuid
import socket
import time
import json
import signal

from kunai.cluster import Cluster
from kunai.log import cprint, logger
from kunai.version import VERSION

REDIRECT_TO = getattr(os, "devnull", "/dev/null")

# Main class for launching the daemon
class Launcher(object):
    def __init__(self, lock_path='', debug_path=''):
        self.lock_path = lock_path
        self.debug_path = debug_path
        
    
    def change_to_workdir(self):
        if os.path.exists('/tmp'):
            try:
                os.chdir('/tmp')
            except Exception, e:
                raise Exception('Invalid working directory /tmp')
    

    def unlink(self):
        logger.info("Unlinking lock file %s" % self.lock_path)
        try:
            os.unlink(self.lock_path)
        except Exception, e:
            logger.error("Got an error unlinking our pidfile: %s" % (e))



    def open_pidfile(self, write=False):
        try:
            p = os.path.abspath(self.lock_path)
            logger.debug("Opening pid file: %s" % p)
            # Windows do not manage the rw+ mode, so we must open in read mode first, then reopen it write mode...
            if not write and os.path.exists(p):
                self.fpid = open(p, 'r+')
            else:  # If it doesn't exist too, we create it as void
                self.fpid = open(p, 'w+')
        except Exception as err:
            raise Exception('Cannot open pid file: %s' % err)


    # Check (in pidfile) if there isn't already a daemon running. If yes and do_replace: kill it.
    # Keep in self.fpid the File object to the pidfile. Will be used by writepid.
    def check_parallel_run(self):
        # TODO: other daemon run on nt
        if os.name == 'nt':
            logger.warning("The parallel daemon check is not available on nt")
            self.open_pidfile(write=True)
            return

        # First open the pid file in open mode
        self.open_pidfile()
        try:
            buf = self.fpid.readline().strip(' \r\n')
            if not buf: # pid file was void, cool
                return
            pid = int(buf)
        except Exception as err:
            logger.info("Stale pidfile exists at %s (%s). Reusing it." % (err, self.lock_path))
            return

        try:
            os.kill(pid, 0)
        except Exception as err: # consider any exception as a stale pidfile.
            # this includes :
            #  * PermissionError when a process with same pid exists but is executed by another user.
            #  * ProcessLookupError: [Errno 3] No such process.
            logger.info("Stale pidfile exists (%s), Reusing it." % err)
            return

        logger.error("Valid previous daemon exists (pid=%s) Exiting." % pid)
        raise SystemExit(2)
    

    def write_pid(self, pid=None):
        if pid is None:
            pid = os.getpid()
        self.fpid.seek(0)
        self.fpid.truncate()
        self.fpid.write("%d" % (pid))
        self.fpid.close()
        del self.fpid  ## no longer needed



    # Go in "daemon" mode: redirect stdout/err,
    # chdir, umask, fork-setsid-fork-writepid
    def daemonize(self):
        logger.debug("Redirecting stdout and stderr as necessary..")
        if self.debug_path:
            fdtemp = os.open(self.debug_path, os.O_CREAT | os.O_WRONLY | os.O_TRUNC)
        else:
            fdtemp = os.open(REDIRECT_TO, os.O_RDWR)

        os.dup2(fdtemp, 1)  # standard output (1)
        os.dup2(fdtemp, 2)  # standard error (2)
        
        # Now the fork/setsid/fork..
        try:
            pid = os.fork()
        except OSError, e:
            s = "%s [%d]" % (e.strerror, e.errno)
            logger.error(s)
            raise Exception, s
        
        if pid != 0:
            # In the father: we check if our child exit correctly
            # it has to write the pid of our future little child..
            def do_exit(sig, frame):
                logger.error("Timeout waiting child while it should have quickly returned ; something weird happened")
                os.kill(pid, 9)
                sys.exit(2)
            # wait the child process to check its return status:
            signal.signal(signal.SIGALRM, do_exit)
            signal.alarm(3)  # forking & writing a pid in a file should be rather quick..
            # if it's not then something wrong can already be on the way so let's wait max 3 secs here.
            pid, status = os.waitpid(pid, 0)
            if status != 0:
                logger.error("Something weird happened with/during second fork: status=", status)
                os._exit(2)
            # In all case we will have to return
            os._exit(0)
        
        # halfway to daemonize..
        os.setsid()
        try:
            pid = os.fork()
        except OSError, e:
            raise Exception, "%s [%d]" % (e.strerror, e.errno)
        if pid != 0:
            # we are the last step and the real daemon is actually correctly created at least.
            # we have still the last responsibility to write the pid of the daemon itself.
            self.write_pid(pid)
            os._exit(0) # <-- this was the son, the real daemon is the son-son
        
        self.fpid.close()
        del self.fpid
        self.pid = os.getpid()
        logger.info("Daemonization done: pid=%d" % self.pid)
        

    def do_daemon_init_and_start(self, is_daemon=False):
        self.change_to_workdir()
        self.check_parallel_run()

        # Force the debug level if the daemon is said to start with such level
        if self.debug_path:
            logger.setLevel('DEBUG')
        
        # If daemon fork() until we reach the final step
        if is_daemon:
            self.daemonize()
        else:
            self.write_pid()

        # Here only the son-son reach this part :)
        
        
    # Main locking function, will LOCK here until the daemon is dead/killed/whatever
    def main(self):
        c = Cluster()
        cprint('Linking services and checks', color='green')    
        c.link_services()
        c.link_checks()
        cprint('Launching listeners', color='green')    
        c.launch_listeners()
        cprint('Joining seeds nodes', color='green')    
        c.join()
        cprint('Starting check, collector and generator threads', color='green')    
        c.launch_check_thread()
        c.launch_collector_thread()
        c.launch_generator_thread()
    
        if 'kv' in c.tags:
            c.launch_replication_backlog_thread()
            c.launch_replication_first_sync_thread()
        if 'ts' in c.tags:
            c.start_ts_listener()

        # Blocking function here
        c.main()
        
    



import httplib # Used only for handling httplib.HTTPException (case #26701)
import os
import platform
import re
import urllib
import urllib2
import traceback
import time
from StringIO import StringIO


from kunai.log import logger
from kunai.collector import Collector


class Mysql(Collector):
    def launch(self):
        logger.debug('getMySQLStatus: start')

        if 'MySQLServer' in self.config and 'MySQLUser' in self.config and self.config['MySQLServer'] != '' and self.config['MySQLUser'] != '':

            logger.debug('getMySQLStatus: config')

            # Try import MySQLdb - http://sourceforge.net/projects/mysql-python/files/
            try:
                import MySQLdb
            except ImportError, e:
                logger.error('getMySQLStatus: unable to import MySQLdb')
                return False

            if 'MySQLPort' not in self.config:
                self.config['MySQLPort'] = 3306

            if 'MySQLSocket' not in self.config:
                # Connect
                try:
                    db = MySQLdb.connect(host=self.config['MySQLServer'], user=self.config['MySQLUser'], passwd=self.config['MySQLPass'], port=int(self.config['MySQLPort']))
                except MySQLdb.OperationalError, message:
                    logger.error('getMySQLStatus: MySQL connection error (server): %s', message)
                    return False
            else:
                # Connect
                try:
                    db = MySQLdb.connect(host='localhost', user=self.config['MySQLUser'], passwd=self.config['MySQLPass'], port=int(self.config['MySQLPort']), unix_socket=self.config['MySQLSocket'])

                except MySQLdb.OperationalError, message:
                    logger.error('getMySQLStatus: MySQL connection error (socket): %s', message)
                    return False

            logger.debug('getMySQLStatus: connected')

            # Get MySQL version
            if self.mysqlVersion == None:
                logger.debug('getMySQLStatus: mysqlVersion unset storing for first time')
                try:
                    cursor = db.cursor()
                    cursor.execute('SELECT VERSION()')
                    result = cursor.fetchone()
                except MySQLdb.OperationalError, message:
                    logger.error('getMySQLStatus: MySQL query error when getting version: %s', message)

                version = result[0].split('-') # Case 31237. Might include a description e.g. 4.1.26-log. See http://dev.mysql.com/doc/refman/4.1/en/information-functions.html#function_version
                version = version[0].split('.')
                self.mysqlVersion = []

                # Make sure the version is only an int. Case 31647
                for string in version:
                    number = re.match('([0-9]+)', string)
                    number = number.group(0)
                    self.mysqlVersion.append(number)

            logger.debug('getMySQLStatus: getting Connections')

            # Connections
            try:
                cursor = db.cursor()
                cursor.execute('SHOW STATUS LIKE "Connections"')
                result = cursor.fetchone()

            except MySQLdb.OperationalError, message:
                logger.error('getMySQLStatus: MySQL query error when getting Connections = %s', message)

            if self.mysqlConnectionsStore == None:
                logger.debug('getMySQLStatus: mysqlConnectionsStore unset storing for first time')
                self.mysqlConnectionsStore = result[1]
                connections = 0
            else:
                logger.debug('getMySQLStatus: mysqlConnectionsStore set so calculating')
                logger.debug('getMySQLStatus: self.mysqlConnectionsStore = %s', self.mysqlConnectionsStore)
                logger.debug('getMySQLStatus: result = %s', result[1])
                connections = float(float(result[1]) - float(self.mysqlConnectionsStore)) / 60
                self.mysqlConnectionsStore = result[1]

            logger.debug('getMySQLStatus: connections  = %s', connections)
            logger.debug('getMySQLStatus: getting Connections - done')
            logger.debug('getMySQLStatus: getting Created_tmp_disk_tables')

            # Created_tmp_disk_tables

            # Determine query depending on version. For 5.02 and above we need the GLOBAL keyword (case 31015)
            if int(self.mysqlVersion[0]) >= 5 and int(self.mysqlVersion[2]) >= 2:
                query = 'SHOW GLOBAL STATUS LIKE "Created_tmp_disk_tables"'
            else:
                query = 'SHOW STATUS LIKE "Created_tmp_disk_tables"'

            try:
                cursor = db.cursor()
                cursor.execute(query)
                result = cursor.fetchone()
            except MySQLdb.OperationalError, message:
                logger.error('getMySQLStatus: MySQL query error when getting Created_tmp_disk_tables = %s', message)

            createdTmpDiskTables = float(result[1])

            logger.debug('getMySQLStatus: createdTmpDiskTables = %s', createdTmpDiskTables)
            logger.debug('getMySQLStatus: getting Created_tmp_disk_tables - done')
            logger.debug('getMySQLStatus: getting Max_used_connections')

            # Max_used_connections
            try:
                cursor = db.cursor()
                cursor.execute('SHOW STATUS LIKE "Max_used_connections"')
                result = cursor.fetchone()
            except MySQLdb.OperationalError, message:
                logger.error('getMySQLStatus: MySQL query error when getting Max_used_connections = %s', message)

            maxUsedConnections = result[1]
            logger.debug('getMySQLStatus: maxUsedConnections = %s', createdTmpDiskTables)
            logger.debug('getMySQLStatus: getting Max_used_connections - done')
            logger.debug('getMySQLStatus: getting Open_files')

            # Open_files
            try:
                cursor = db.cursor()
                cursor.execute('SHOW STATUS LIKE "Open_files"')
                result = cursor.fetchone()
            except MySQLdb.OperationalError, message:
                logger.error('getMySQLStatus: MySQL query error when getting Open_files = %s', message)

            openFiles = result[1]

            logger.debug('getMySQLStatus: openFiles = %s', openFiles)
            logger.debug('getMySQLStatus: getting Open_files - done')
            logger.debug('getMySQLStatus: getting Slow_queries')

            # Slow_queries

            # Determine query depending on version. For 5.02 and above we need the GLOBAL keyword (case 31015)
            if int(self.mysqlVersion[0]) >= 5 and int(self.mysqlVersion[2]) >= 2:
                query = 'SHOW GLOBAL STATUS LIKE "Slow_queries"'

            else:
                query = 'SHOW STATUS LIKE "Slow_queries"'

            try:
                cursor = db.cursor()
                cursor.execute(query)
                result = cursor.fetchone()
            except MySQLdb.OperationalError, message:
                logger.error('getMySQLStatus: MySQL query error when getting Slow_queries = %s', message)

            if self.mysqlSlowQueriesStore == None:
                logger.debug('getMySQLStatus: mysqlSlowQueriesStore unset so storing for first time')
                self.mysqlSlowQueriesStore = result[1]
                slowQueries = 0

            else:
                logger.debug('getMySQLStatus: mysqlSlowQueriesStore set so calculating')
                logger.debug('getMySQLStatus: self.mysqlSlowQueriesStore = %s', self.mysqlSlowQueriesStore)
                logger.debug('getMySQLStatus: result = %s', result[1])

                slowQueries = float(float(result[1]) - float(self.mysqlSlowQueriesStore)) / 60

                self.mysqlSlowQueriesStore = result[1]

            logger.debug('getMySQLStatus: slowQueries = %s', slowQueries)
            logger.debug('getMySQLStatus: getting Slow_queries - done')
            logger.debug('getMySQLStatus: getting Table_locks_waited')

            # Table_locks_waited
            try:
                cursor = db.cursor()
                cursor.execute('SHOW STATUS LIKE "Table_locks_waited"')
                result = cursor.fetchone()
            except MySQLdb.OperationalError, message:
                logger.error('getMySQLStatus: MySQL query error when getting Table_locks_waited = %s', message)

            tableLocksWaited = float(result[1])

            logger.debug('getMySQLStatus: tableLocksWaited  = %s', tableLocksWaited)
            logger.debug('getMySQLStatus: getting Table_locks_waited - done')
            logger.debug('getMySQLStatus: getting Threads_connected')

            # Threads_connected
            try:
                cursor = db.cursor()
                cursor.execute('SHOW STATUS LIKE "Threads_connected"')
                result = cursor.fetchone()
            except MySQLdb.OperationalError, message:
                logger.error('getMySQLStatus: MySQL query error when getting Threads_connected = %s', message)

            threadsConnected = result[1]

            logger.debug('getMySQLStatus: threadsConnected = %s', threadsConnected)
            logger.debug('getMySQLStatus: getting Threads_connected - done')
            logger.debug('getMySQLStatus: getting Seconds_Behind_Master')

            if 'MySQLNoRepl' not in self.config:
                # Seconds_Behind_Master
                try:
                    cursor = db.cursor(MySQLdb.cursors.DictCursor)
                    cursor.execute('SHOW SLAVE STATUS')
                    result = cursor.fetchone()
                except MySQLdb.OperationalError, message:
                    logger.error('getMySQLStatus: MySQL query error when getting SHOW SLAVE STATUS = %s', message)
                    result = None

                if result != None:
                    try:
                        secondsBehindMaster = result['Seconds_Behind_Master']
                        logger.debug('getMySQLStatus: secondsBehindMaster = %s', secondsBehindMaster)

                    except IndexError, e:
                        secondsBehindMaster = None
                        logger.debug('getMySQLStatus: secondsBehindMaster empty. %s', e)

                else:
                    secondsBehindMaster = None
                    logger.debug('getMySQLStatus: secondsBehindMaster empty. Result = None.')

                logger.debug('getMySQLStatus: getting Seconds_Behind_Master - done')

            return {'connections' : connections, 'createdTmpDiskTables' : createdTmpDiskTables, 'maxUsedConnections' : maxUsedConnections, 'openFiles' : openFiles, 'slowQueries' : slowQueries, 'tableLocksWaited' : tableLocksWaited, 'threadsConnected' : threadsConnected, 'secondsBehindMaster' : secondsBehindMaster}

        else:
            logger.debug('getMySQLStatus: config not set')
            return False



import os
import sys
import imp
import glob

from kunai.collector import Collector
from kunai.log import logger


def get_collectors(self):
    collector_dir = os.path.dirname(__file__)
    p = collector_dir+'/*py'
    print "LOAD", p
    collector_files = glob.glob(p)
    for f in collector_files:
        fname = os.path.splitext(os.path.basename(f))[0]
        try:
            m = imp.load_source('collector%s' % fname, f)
        except Exception, exp:
            print "COLLECTOR LOAD FAIL", exp
            continue

    collector_clss = Collector.get_sub_class()
    for ccls in collector_clss:
        # skip base module Collector
        if ccls == Collector:
            continue
        # Maybe this collector is already loaded
        if ccls in self.collectors:
            continue
        self.load_collector(ccls)




import json
from websocketserver import WebSocket, SimpleWebSocketServer




class WebExporter(WebSocket):
    def handleMessage(self):
        if self.data is None:
            self.data = ''

    def handleConnected(self):
        print self.address, 'connected'

    def handleClose(self):
        print self.address, 'closed'


class WebSocketBackend(object):
    def __init__(self, clust):
        self.clust = clust
        port = clust.websocket.get('port', 6769)
        self.server = SimpleWebSocketServer(clust.addr, port, WebExporter)


    def run(self):
	self.server.serveforever()


    def get_info(self):
        return {'nb_connexions': len(self.server.connections)}
    

    def send_all(self, o):
        try:
            msg = json.dumps(o)
        except ValueError:
            print "BAD MESSAGE"
            return
        for client in self.server.connections.itervalues():
            print "SENDING"*100
            try:
                client.sendMessage(msg)
            except Exception as n:
                print n
            print "SENT DONE"*100
        



#!/usr/bin/env python
# Copyright (C) 2014:
#    Gabes Jean, naparuba@gmail.com

import copy
import time
import threading
from Queue import Queue,Empty

from kunai_test import *
from kunai.raft import RaftNode


class TestRaft(KunaiTest):
    def setUp(self):
        pass

    def create(self, N=3):
        self.nodes = [{'node':RaftNode(i), 'queue': Queue()} for i in range(N)]
        

    def compute_stats(self):
        self.stats = {}        
        for d in self.nodes:
            n = d['node']
            s = n.state
            if s not in self.stats:
                self.stats[s] = 0
            self.stats[s] += 1

            
    def count(self, state):
        # hummering the stats so we are up to date
        self.compute_stats()
        return self.stats.get(state, 0)

    
    def launch(self):
        #nodes = [{'node':RaftNode(i), 'queue': Queue()} for i in range(N)]

        self.threads = []
        for d in self.nodes:
            n = d['node']
            q = d['queue']
            t = threading.Thread(None, target=n.main, name='node-%d' % n.i, args=(q, self.nodes))
            t.daemon = True
            t.start()
            self.threads.append(t)

        self.start = time.time()

        return

    
        for t in threads:
            t.join()

        # did we got a leader?
        print "RESULT FOR", LOOP
        leader = None
        max_vote = 0
        for d in nodes:
            n = d['node']
            max_vote = max(max_vote, n.nb_vote)
            if n.state == 'leader':
                if leader != None:
                    print "WE GOT 2 LEADER, WTF DID YOU DID JEAN?????"
                    sys.exit("JEAN HOW CAN YOU BREAK SUCH AN ALGO?")

                print "GOT A LEADER", n.i, 'with ', n.nb_vote, "LOOP", LOOP
                leader = n

        print "Candidate density::", LOOP, 300*(2**LOOP) / float(N), "ms", "& number of candidate in this loop (%d)" % LOOP, len([d for d in nodes if d['node'].state in ('candidate', 'leader')])
        if leader is not None:
            print "Good job jim", "LOOP", LOOP
            sys.exit(0)

        print "No leader, max vote is", max_vote


    def stop(self):
        for d in self.nodes:
            n = d['node']
            n.stop()
        for t in self.threads:
            t.join(2)


    # Create N nodes with their own thread, and wait some seconds 
    def create_and_wait(self, N=3, wait=3):
        self.create(N)
        self.launch()
        time.sleep(wait)


    def get_leader(self):
        for d in self.nodes:
            n = d['node']
            if n.state == 'leader':
                return n


    def get_all_state(self, state):
        res = []
        for d in self.nodes:
            n = d['node']
            if n.state == state:
                res.append(n)
        return res
        

############################### TESTS        
        
    def test_raft_simple_leader_election(self):
        self.create_and_wait(N=3, wait=3)

        self.assert_(self.count('leader') == 1)        
        # always clean before exiting a test
        self.stop()
        

        
    # Try with far more nodes
    def test_raft_large_leader_election(self):
        N = 50
        W = 10 # for very slow computing like travis?
        self.create_and_wait(N=N, wait=W)

        print "Looking if we really got a leader, and only one"        
        self.assert_(self.count('leader') == 1)

        # and N-1 followers
        nb_followers = self.count('follower')
        print "NB followers", nb_followers
        print self.stats
        self.assert_(nb_followers == N-1)
        
        # always clean before exiting a test
        self.stop()
        
        
    
    # What is we kill some not leader nodes?
    def test_raft_kill_no_leader(self):
        N = 10
        W = 3
        self.create_and_wait(N=N, wait=W)
        print "Looking if we really got a leader, and only one"        
        self.assert_(self.count('leader') == 1)
        self.assert_(self.count('follower') == N-1)

        followers = self.get_all_state('follower')
        print followers

        NB_kill = N/2
        killed = []
        for i in range(NB_kill):
            n = followers[i]
            n.stop()
            killed.append(n)
        # wait for node to accept it
        time.sleep(1)
        
        states = [n.state for n in killed]
        # look if all are dead
        self.assert_(states.count('leaved') == NB_kill)

        # there should be still one leader
        self.assert_(self.count('leader') == 1)
        # but there are less followers now
        self.assert_(self.count('follower') == N-1-NB_kill)
        
        
        # always clean before exiting a test
        self.stop()
        

        
        
if __name__ == '__main__':
    unittest.main()



import re
import socket



pattern = r"((25[0-5]|2[0-4][0-9]|[01]?[0-9][0-9]?)([ (\[]?(\.|dot)[ )\]]?(25[0-5]|2[0-4][0-9]|[01]?[0-9][0-9]?)){3})"
ipv4pattern = re.compile(pattern)

class DNSQuery:
   def __init__(self, data):
      self.data = data
      self.domain = ''
      
      t = (ord(data[2]) >> 3) & 15   # Opcode bits
      if t == 0:                     # Standard query
         ini = 12
         lon = ord(data[ini])
         while lon != 0:
            self.domain += data[ini+1:ini+lon+1]+'.'
            ini += lon+1
            lon = ord(data[ini])

      
   def _get_size_hex(self, nb):
      nb = min(nb, 256*256)
      d,r = divmod(nb, 256)
      s = chr(d)+chr(r)
      return s

    
   # We look in the nodes for the good tag
   def lookup_for_nodes(self, nodes, dom):
      print "DNS LOOKING ", self.domain, "inside domain", dom
      if not self.domain.endswith(dom):
          return []
      search = self.domain[:-len(dom)]
      # split into sname.service.datacenter
      print "DNS lookup for search", search
      elts = search.split('.', 2)
      if len(elts) != 3:
         print "DNS bad query", search
         return []
      dc = elts[2]
      _type = elts[1]
      tag = elts[0]
      r = []
      for n in nodes.values():
         # skip non alive nodes
         if n['state'] != 'alive':
            continue
         if tag in n['tags']:
            services = n.get('services', {})
            state_id = 0
            if tag in services:
               service = services[tag]
               state_id = service.get('state_id')
            print "DNS state_id", state_id
            if state_id == 0:
               addr = n['addr']
               # If already an ip, add it
               if ipv4pattern.match(addr):
                  r.append(addr)
               else: # else try to resolv it first
                  try:
                     addr = socket.gethostbyname(addr)
                     r.append(addr)
                  except socket.gaierror: # not found
                     print 'DNS cannot find the hotname ip', addr
                     # skip this node
      
      print "DNS R:", r
      return r


   def response(self, r):
      packet = ''
      print "DNS DOM", self.domain
      nb = len(r)
      if self.domain:
         packet += self.data[:2] + "\x81\x80"
         packet += self.data[4:6] + self._get_size_hex(nb) + '\x00\x00\x00\x00'   # Questions and Answers Counts
         packet += self.data[12:]                                         # Original Domain Name Question

         for ip in r:
            packet += '\xc0\x0c'                                 # Pointer to domain name
            packet += '\x00\x01\x00\x01\x00\x00\x00\x3c\x00\x04' # Response type, ttl and resource data length -> 4 bytes
            packet += str.join('',map(lambda x: chr(int(x)), ip.split('.'))) # 4bytes of IP
      
      print "DNS RETURNing", len(packet), len(r)
      return packet




# Strange file to to fine tuning. Gettimeofday can be CPU (%sys) consuming
# if call 100K/s, so if you need only a "somewhat now", call, like with a 1/100s
# precision, you can call this lib object NOW that will be updated 100time a sec

import time
from kunai.threadmgr import threader

class QuickNow(object):
    def __init__(self):
        self.now = int(time.time())


    def do_thread(self):
        while True:
            self.now = int(time.time())
            time.sleep(0.01) # update each 10ms
    
    
    def launch(self):
        threader.create_and_launch(self.do_thread, name='now-thread')


NOW = QuickNow()



import httplib # Used only for handling httplib.HTTPException (case #26701)
import os
import sys
import platform
import re
import urllib
import urllib2
import traceback
import time
from StringIO import StringIO


from kunai.log import logger
from kunai.collector import Collector


class NetworkTraffic(Collector):
    def launch(self):
        logger.debug('getNetworkTraffic: start')

        if sys.platform == 'linux2':
            logger.debug('getNetworkTraffic: linux2')

            try:
                logger.debug('getNetworkTraffic: attempting open')
                proc = open('/proc/net/dev', 'r')
                lines = proc.readlines()
                proc.close()

            except IOError, e:
                logger.error('getNetworkTraffic: exception = %s', e)
                return False

            logger.debug('getNetworkTraffic: open success, parsing')

            columnLine = lines[1]
            _, receiveCols , transmitCols = columnLine.split('|')
            receiveCols = map(lambda a:'recv_' + a, receiveCols.split())
            transmitCols = map(lambda a:'trans_' + a, transmitCols.split())

            cols = receiveCols + transmitCols

            logger.debug('getNetworkTraffic: parsing, looping')

            faces = {}
            for line in lines[2:]:
                if line.find(':') < 0: continue
                face, data = line.split(':')
                faceData = dict(zip(cols, data.split()))
                faces[face] = faceData

            logger.debug('getNetworkTraffic: parsed, looping')

            interfaces = {}

            # Now loop through each interface
            for face in faces:
                key = face.strip()

                # We need to work out the traffic since the last check so first time we store the current value
                # then the next time we can calculate the difference
                try:
                    if key in self.networkTrafficStore:
                        interfaces[key] = {}
                        interfaces[key]['recv_bytes'] = long(faces[face]['recv_bytes']) - long(self.networkTrafficStore[key]['recv_bytes'])
                        interfaces[key]['trans_bytes'] = long(faces[face]['trans_bytes']) - long(self.networkTrafficStore[key]['trans_bytes'])

                        if interfaces[key]['recv_bytes'] < 0:
                            interfaces[key]['recv_bytes'] = long(faces[face]['recv_bytes'])

                        if interfaces[key]['trans_bytes'] < 0:
                            interfaces[key]['trans_bytes'] = long(faces[face]['trans_bytes'])

                        # Get the traffic
                        interfaces[key]['recv_bytes/s'] = interfaces[key]['recv_bytes'] / 10
                        interfaces[key]['trans_bytes/s'] = interfaces[key]['trans_bytes'] / 10

                        interfaces[key]['recv_bytes'] = str(interfaces[key]['recv_bytes'])
                        interfaces[key]['trans_bytes'] = str(interfaces[key]['trans_bytes'])


                        # And update the stored value to subtract next time round
                        self.networkTrafficStore[key]['recv_bytes'] = faces[face]['recv_bytes']
                        self.networkTrafficStore[key]['trans_bytes'] = faces[face]['trans_bytes']

                    else:
                        self.networkTrafficStore[key] = {}
                        self.networkTrafficStore[key]['recv_bytes'] = faces[face]['recv_bytes']
                        self.networkTrafficStore[key]['trans_bytes'] = faces[face]['trans_bytes']

                    # Logging
                    logger.debug('getNetworkTraffic: %s = %s', key, self.networkTrafficStore[key]['recv_bytes'])
                    logger.debug('getNetworkTraffic: %s = %s', key, self.networkTrafficStore[key]['trans_bytes'])

                except KeyError, ex:
                    logger.error('getNetworkTraffic: no data for %s', key)

                except ValueError, ex:
                    logger.error('getNetworkTraffic: invalid data for %s', key)

            logger.debug('getNetworkTraffic: completed, returning')

            return interfaces

        elif sys.platform.find('freebsd') != -1:
            logger.debug('getNetworkTraffic: freebsd')

            try:
                try:
                    logger.debug('getNetworkTraffic: attempting Popen (netstat)')

                    proc = subprocess.Popen(['netstat', '-nbid'], stdout=subprocess.PIPE, close_fds=True)
                    netstat = proc.communicate()[0]

                    if int(pythonVersion[1]) >= 6:
                        try:
                            proc.kill()
                        except Exception, e:
                            logger.debug('Process already terminated')

                except Exception, e:
                    logger.error('getNetworkTraffic: exception = %s', traceback.format_exc())

                    return False
            finally:
                if int(pythonVersion[1]) >= 6:
                    try:
                        proc.kill()
                    except Exception, e:
                        logger.debug('Process already terminated')

            logger.debug('getNetworkTraffic: open success, parsing')

            lines = netstat.split('\n')

            # Loop over available data for each inteface
            faces = {}
            rxKey = None
            txKey = None

            for line in lines:
                logger.debug('getNetworkTraffic: %s', line)

                line = re.split(r'\s+', line)

                # Figure out which index we need
                if rxKey == None and txKey == None:
                    for k, part in enumerate(line):
                        logger.debug('getNetworkTraffic: looping parts (%s)', part)

                        if part == 'Ibytes':
                            rxKey = k
                            logger.debug('getNetworkTraffic: found rxKey = %s', k)
                        elif part == 'Obytes':
                            txKey = k
                            logger.debug('getNetworkTraffic: found txKey = %s', k)

                else:
                    if line[0] not in faces:
                        try:
                            logger.debug('getNetworkTraffic: parsing (rx: %s = %s / tx: %s = %s)', rxKey, line[rxKey], txKey, line[txKey])
                            faceData = {'recv_bytes': line[rxKey], 'trans_bytes': line[txKey]}

                            face = line[0]
                            faces[face] = faceData
                        except IndexError, e:
                            continue

            logger.debug('getNetworkTraffic: parsed, looping')

            interfaces = {}

            # Now loop through each interface
            for face in faces:
                key = face.strip()

                try:
                    # We need to work out the traffic since the last check so first time we store the current value
                    # then the next time we can calculate the difference
                    if key in self.networkTrafficStore:
                        interfaces[key] = {}
                        interfaces[key]['recv_bytes'] = long(faces[face]['recv_bytes']) - long(self.networkTrafficStore[key]['recv_bytes'])
                        interfaces[key]['trans_bytes'] = long(faces[face]['trans_bytes']) - long(self.networkTrafficStore[key]['trans_bytes'])

                        interfaces[key]['recv_bytes'] = str(interfaces[key]['recv_bytes'])
                        interfaces[key]['trans_bytes'] = str(interfaces[key]['trans_bytes'])

                        if interfaces[key]['recv_bytes'] < 0:
                            interfaces[key]['recv_bytes'] = long(faces[face]['recv_bytes'])

                        if interfaces[key]['trans_bytes'] < 0:
                            interfaces[key]['trans_bytes'] = long(faces[face]['trans_bytes'])



                        # And update the stored value to subtract next time round
                        self.networkTrafficStore[key]['recv_bytes'] = faces[face]['recv_bytes']
                        self.networkTrafficStore[key]['trans_bytes'] = faces[face]['trans_bytes']

                    else:
                        self.networkTrafficStore[key] = {}
                        self.networkTrafficStore[key]['recv_bytes'] = faces[face]['recv_bytes']
                        self.networkTrafficStore[key]['trans_bytes'] = faces[face]['trans_bytes']

                except KeyError, ex:
                    logger.error('getNetworkTraffic: no data for %s', key)

                except ValueError, ex:
                    logger.error('getNetworkTraffic: invalid data for %s', key)

            logger.debug('getNetworkTraffic: completed, returning')

            return interfaces

        else:
            logger.debug('getNetworkTraffic: other platform, returning')

            return False



#!/usr/bin/env python
# Copyright (C) 2014:
#    Gabes Jean, naparuba@gmail.com

import copy
import time
import threading
from kunai_test import *

from kunai.pubsub import pubsub



class TestPubSub(KunaiTest):
    def setUp(self):
        self.f1_raised = False
        self.f2_raised = False
        
        self.f3_raised = False        
        self.f3_value = None

        
    def f1(self):
        self.f1_raised = True

        
    def f2(self):
        self.f2_raised = True

    def f3(self, value_set=''):
        self.f3_raised = True
        if value_set:
            self.f3_value = value_set


        
    def test_pubsub(self):
        # Not registerred, so won't trigger it
        pubsub.pub('main')
        self.assert_(self.f1_raised == False)

        pubsub.sub('main', self.f1)
        pubsub.pub('main')
        self.assert_(self.f1_raised == True)

        # hard reset the bool and do with a two this time
        self.f1_raised = False
        pubsub.sub('main', self.f1)
        pubsub.sub('main', self.f2)
        pubsub.pub('main')
        self.assert_(self.f1_raised == True)
        self.assert_(self.f2_raised == True)

        # Now with params
        pubsub.sub('params', self.f3)
        pubsub.pub('params', value_set='set')
        self.assert_(self.f3_raised == True)
        self.assert_(self.f3_value == 'set')

if __name__ == '__main__':
    unittest.main()



#!/usr/bin/python

# -*- coding: utf-8 -*-


#==> need to rewrite or at least have a common lib with shinken or nagios

import re


perfdata_split_pattern = re.compile('([^=]+=\S+)')
metric_pattern = re.compile('^([^=]+)=([\d\.\-\+eE]+)([\w\/%]*);?([\d\.\-\+eE:~@]+)?;?([\d\.\-\+eE:~@]+)?;?([\d\.\-\+eE]+)?;?([\d\.\-\+eE]+)?;?\s*')


def to_best_int_float(val):
    i = int(float(val))
    f = float(val)
    # If the f is a .0 value,
    # best match is int
    if i == f:
        return i
    return f


# If we can return an int or a float, or None
# if we can't
def guess_int_or_float(val):
    try:
        return to_best_int_float(val)
    except Exception, exp:
        return None


# Class for one metric of a perf_data
class Metric:
    def __init__(self, s):
        self.name = self.value = self.uom = self.warning = self.critical = self.min = self.max = None
        s = s.strip()
        r = metric_pattern.match(s)
        if r:
            # Get the name but remove all ' in it
            self.name = r.group(1).replace("'", "")
            self.value = guess_int_or_float(r.group(2))
            self.uom = r.group(3)
            self.warning = guess_int_or_float(r.group(4))
            self.critical = guess_int_or_float(r.group(5))
            self.min = guess_int_or_float(r.group(6))
            self.max = guess_int_or_float(r.group(7))
            if self.uom == '%':
                self.min = 0
                self.max = 100

    def __str__(self):
        s = "%s=%s%s" % (self.name, self.value, self.uom)
        if self.warning:
            s = s + ";%s" % (self.warning)
        if self.critical:
            s = s + ";%s" % (self.critical)
        return s


class PerfDatas:
    def __init__(self, s):
        s = s or ''
        elts = perfdata_split_pattern.findall(s)
        elts = [e for e in elts if e != '']
        self.metrics = {}
        for e in elts:
            m = Metric(e)
            if m.name is not None:
                self.metrics[m.name] = m

    def __iter__(self):
        return self.metrics.itervalues()

    def __len__(self):
        return len(self.metrics)

    def __getitem__(self, key):
        return self.metrics[key]

    def __contains__(self, key):
        return key in self.metrics

